/*
 * This file contains the user specified input for the integrator settings, more detailed explanations for all 
 * of these settings can be found in the Cuba manual.
 */

#include <Auxiliaries.h>

/*
 * Divonne uses a three-step integration. 
 * First, the integration domain is partitioned into areas of roughly equal "spread", which is the area of 
 * integration multiplied with the difference between integrand function maximum and minimum in the subregion.
 * In the second step the interior of all subregions is sampled for the final result.
 * Third, if the deviation between the estimated result from the partitioning stage and the sampling stage 
 * fails a chi squared test, the subregion in question can be subdivided again.
 *
 * The parameters below define the methods used for partitioning and sampling, as well as a few other auxiliary settings.
 */

/* This block of parameters governs the desired accuracy of the integration. 'epsrelI' describe the relative 
 * accuracy of the I-dimensional integration, i.e. epsrel2=0.02 runs the integration until for an integration 
 * result RESULT, the error as delivered by the integrator lies below 0.02*RESULT. 
 * A setting for absolute errors exists as well, but is maintained only for compatibility issues, the relevant 
 * settings can be found at the end of this file.
 * 
 * Note that this demand on accuracy already applies at the partitioning stage, so that the final result typically 
 * has a lower error estimate than demanded. The standard settings for the sampling below add typically 2 orders of 
 * magnitude, so that epsrel2=0.001 delivers final results that show ERROR=0.00001*RESULT.
 */
const double epsrel2=0.001;	//the default is 0.001
const double epsrel3=0.001;	//the default is 0.001
const double epsrel4=0.01;	//the default is 0.01
const double epsrel5=0.1;	//the default is 0.1

/* These groups of three determine the integration strategies for the partitioning ("key1X"), sampling ("key2X")
 * and refinement ("key3X") stages for the X-dim integration, respectively. The possible settings are as follows:
 * 
 * For "key1X" and "key2X", positive numbers correspond generally to sampling using Korobov quasi-random numbers.
 * Note that there are only ~16000 Korobov random numbers available to Cuba. Setting a key variable to a value >16000 
 * is therefore not better than setting it to 16000. 
 * Certain numbers are special: 7,9 and 13(for "key2X" only) correspond to using Cuhre-type deterministic cubature 
 * rules of degree 7,9 and 13, respectively. 
 * Negative numbers correspond to Sobol or pseudo-random number sampling. In this case, the "seed" variable below 
 * sets the source of these numbers.
 *
 * "key2X" has a special feature: If abs(key2X)<40 (and !=7,9 or 13), Divonne uses this value to multiply an 
 * estimate for the number of required samplings to achieve the desired accuracy, which it derives from the 
 * partitioning stage.
 * 
 * "key3X" sets whether the refinement phase should subdivide again in the case of mismatching resuts from partitioning 
 * and sampling (key3X=1) or do nothing (key3X=0), or should sample the region again (without splitting) ("key3X" here 
 * follows the "key2X" system of determining how to sample).
 *
 * The default settings are Cubature-based partitioning using the rules with highest available degree, followed by Korobov 
 * sampling and refinement that allows subdivision.
 */
const int key12=13;	//the default is 13
const int key22=16000;	//the default is 16000
const int key32=1;	//the default is 1

const int key13=9;	//the default is 9 (the Cuba manual suggests 11 might work, but it produces NaN reliably)
const int key23=16000;	//the default is 16000
const int key33=1;	//the default is 1

const int key14=9;	//the default is 9
const int key24=16000;	//the default is 16000
const int key34=1;	//the default is 1

const int key15=9;	//the default is 9
const int key25=16000;	//the default is 16000
const int key35=1;	//the default is 1

/* This setting enforces a certain number of "safety partitioning steps" during the first (partitioning) stage, 
 * to improve the integral estimate by forcing the partitioning to resolve the integrand a bit better.
 * So while typically the integrator would proceed to the sampling stage as soon as the error estimate lies below
 * 'epsrel' times the integral estimate, we can force it to perform 'maxpass' additional partitioning steps. If 
 * during the safety partitioning the integral or error estimate significantly change (i.e. the integrator discovers 
 * more complicated structure), the counter resets and the integrator continues in the partitioning stage.
 */
const int maxpass2=200;	//the default for the precision setting is 200
const int maxpass3=400;	//the default for the precision setting is 400
const int maxpass4=600;	//the default for the precision setting is 600
const int maxpass5=400;	//the default for the precision setting is 400

/* These values set the settings for the chi square test to determine if refinement is necessary.
 * As soon as a subregion exceeds the chi square value set in "maxchisq", it is refined if the contribution to the 
 * total error estimate from this subregion would exceed "mindev"*"total integral error". In other words it first
 * checks if there is reason to refine, and then whether it would change anything.
 */
const double maxchisq=0.5;	//the default is 0.5
const double mindev=0.2;	//the default is 0.2

/* The "border" variable defines an exclusion zone at the boundaries of the integration region inside which no point is
 * directly sampled. Instead, the value of the integrand inside this region is extrapolated by choosing two points outside
 * the border region and extrapolating to the requested point. 
 * The original idea is to avoid endpoint singularities, to which Divonne is sensitive because it tries to resolve extrema 
 * of the integrand really well. For us, there are almost no endpoint singularities by construction (the integrand runs to zero 
 * at most boundaries), but there may be non-trivial behaviour. Accordingly this value should never be set to zero, 
 * values of 10e-7 or 10e-6 are ideal (We run to zero, so linear extrapolation models the integrand adequately and 
 * defuses e.g. instances of x*log(x) at the boundary). 
 *
 * It also serves as a gauge: If you have a measurement function that exhibits large cancellations near the integration 
 * domain boundary, rounding effects due to the finite number of digits in "double" type variables can be relevant. Varying 
 * "border" in these case will result in different final integral and error estimates, and is a reliable gauge that the 
 * multiprecision branch should be used, i.e. programs should be compiled using "make BOOST=1".
 */
const double border=1.0e-8;	//the default is 1.0e-8

/* These variables set upper limits for the number of function evaluations. In general these can be used 
 * to force a program to stop after a certain time, but typically the termination is determinated by epsrel above.
 * For Divonne a special caveat applies: If the number of function evaluations exceeds these limits, the etimated result 
 * from the partitioning stage is returned, not the result of the aborted sampling stage.
 */
const long long int maxeval2=100000000;		//the default is 100000000
const long long int maxeval3=500000000;		//the default is 500000000
const long long int maxeval4=1000000000;	//the default is 1000000000
const long long int maxeval5=5000000000;	//the default is 5000000000

/* If Sobol or pseudo-random sampling is chosen for one of the key values, this determines if Sobol quasi-random numbers
 (seed=0) or pseudo-random numbers (seed= anything else) are used. Details can be found in the Cuba manual.
 */
const int seed=0;

// This variable sets the verbosity of the output. "flags=3" tells you everything, lower values are possible.
const int flags=3;

/* In addition to "epsrel" above, 'epsabsI' in sets the desired absolute accuracy. As for us the integration result 
 * must be multiplied with the correct prefactors, this does not correspond directly to an absolute error estimate 
 * on the bare function.
 * The integrator stops as soon as either epsrel or epsabs are fulfilled. Accordingly, the values are chosen so low 
 * that epsrel always sets the target, as we want to have our accuracy target meaningful and related to the final
 * result for the bare soft function. We list these variables only for compatibility issues.
 */
const double epsabs2=0.000000000001;	//the default is 0.000000000001
const double epsabs3=0.000000000001;	//the default is 0.000000000001
const double epsabs4=0.000000000001;	//the default is 0.000000000001
const double epsabs5=0.000000000001;	//the default is 0.000000000001

/*This file is part of SoftSERVE.
 *
 *   SoftSERVE is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   SoftSERVE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SoftSERVE.  If not, see <http://www.gnu.org/licenses/>.
 */
