/*This file is part of SoftSERVE.
 *
 *   SoftSERVE is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   SoftSERVE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SoftSERVE.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This file contains various auxiliary function definitions, as well as the structures collecting individuals terms 
 * integrated over a fixed number of variables into master formulae to be fed to the integrator.
 *
 * The first definitions contain the derivation of the 1-loop measurement function f as the 1-parton-soft limit of the 
 * 2-loop measurement function F. As the 1-loop version does not know about the variables requiring symmetry splitting 
 * (i.e. the splitting into FA and FB, and FX1 and FX2) it can be derived from any FXi, we choose FA1. 
 *
 * Further, the definitions of the "physical parametrisation" variables (like a,b and tl) in terms of "computing 
 * parametrisation" variables (like z,w,c and s6) are listed here as well. This allows us to let the user specify the 
 * measurement function in the more compact and intuitive physical parametrisation. It also provides a convenient breaking 
 * point for the introduction of multi-precision variables in the case of the "boosted" branch. We know that any potentially  
 * problematic large cancellations occur inside the measurement function, and so it suffices to change any data types used  
 * for measurement function internal computations to multiprecision variables. Accordingly, we use the conversion functions  
 * to translate between "double" and the multi-precision "decimal" types:
 * The integrator provides a "double" type point in the "computing parametrisation"-parametrised integration space. The  
 * reparametrisation functions below convert these to a "decimal" type point in "physical parametrisation"-parametrised  
 * integration space (e.g. decimal b(double c) ). The measurement functions FXi accept these "decimal" type coordinates,  
 * perform internal computations using multi-precision operations, and return a "double" type result to be used in the full,  
 * "double"-valued, integrand function.
 *
 * Additionally, to spare the user the task of having to remember which definition of the angular variables is used and in  
 * which sector Xi of the measurement functions FXi, we define the cosini Ck and Cl of relevant angles betweek \vec{k_T} and  
 * \vec{l_T} and the one allowed distinguished transverse direction \vec{n_T} as functions of the relevant computing  
 * parametrisation variables.
 * 
 *
 * Finally, we define the integrand structures fed to the integrator. These are best described using an example:
 * For the 3-dimensional integration of the CA structure, there are 4 different expressions to be integrated over 3 variables,  
 * depending on which variable out  of (y,u,t5,b) or their rescaled versions is integrated over, in addition to the always 
 * present v/w and tl/s6.
 * Each of these expressions has an eps-expansion, so I get a sum of 4 terms for each appearing eps-order.
 * 
 * Cuba now requires the integrand in a peculiar form: It requires an integer-valued function with two arrays as arguments:  
 * x[nvar] and F[ncomp] (where nvar is the number of integration variables, and ncomp is the number of eps-orders, i.e.  
 * number of different sums of four terms above), and fills the array F with the values of the integrand structures, evaluated  
 * at the point with coordinates x. If this succeeds, the integer it returns is 1. A few auxiliary arguments are included for  
 * these functions, but these are irrelevant for our purposes.
 *
 * The notation is such that "D3IntCA" is the complete 'D'imension '3' 'Int'egrand for the 'CA' structure, and contains all 
 * eps orders, ordered from least divergent (i.e. eps^0) in the first array entry F[0] to the most divergent (for CA eps^-4, 
 * for NF eps^-3) in the last array entry. CA and PNF are combined, with PNF following CA, i.e. F[5] is the least divergent PNf  
 * term, F[8] the most divergent.
 */

#include <Auxiliaries.h>
#include <cmath>
#include <string>
#include <2IntCA.h>
#include <3IntCA.h>
#include <4IntCA.h>
#include <5IntCA.h>
#include <3IntNF.h>
#include <4IntNF.h>
#include <5IntNF.h>
#include <3IntPNF.h>
#include <4IntPNF.h>
#include <5IntPNF.h>
#include <IntOL.h>
#include <IntRV.h>
#include <XFactorCA.h>
#include <Input_Common.h>
#include <iostream>

using namespace std;

double mexp=M(m);

double f(decimal y, decimal t) {
	return FA1(0.5,t,y,t,0.,0.);
}

decimal tl(double s6) {
	return 4*pow(s6,4) - 6*pow(s6,8) + 4*pow(s6,12) - pow(s6,16);
}

decimal b(double c) {
	return pow(c,2.);
}

decimal v(double w) {
	return pow(w,4);
}

decimal t5(double s5) {
	return pow(s5,2);
}

decimal y(double x) {
	return pow(x,mexp);
}

decimal u(double z) {
	return 4*pow(z,2) - 6*pow(z,4) + 4*pow(z,6) - pow(z,8);
}

decimal A(decimal u, decimal v) {
	decimal a;
	a=1.-u+u*v;
	if (a>=1.) {
		a=1.;
	} else if (a<=0.) {
		a=0.;
	}
	return a;
}

decimal T(decimal u, decimal v) {
	decimal t;
	if(u>=1.) {
		t=1.;
	} else if (u<=0) {
		t=0.;
	} else {
		t=pow(u,2)*v/(1.-u+u*v);
	}
	if (isnan(t)) {
		t=1.;
	} else if (t>=1.) {
		t=1.;
	} else if (t<=0.) {
		t=0.;
	}
	return t;
}

decimal Tk1(decimal t, decimal t5, decimal tl) {
	decimal tk1;
	if (t>=1.) {
		tk1=1-tl;
	} else if (t<=0.) {
		tk1=tl;
	} else if (tl>=1.) {
		tk1=1-t;
	} else if (tl<=0.) {
		tk1=t;
	} else {
		tk1=t + tl - 2*t*tl - 2*(t5 - 1)*sqrt((1 - t)*t*(1 - tl)*tl);
	}
	return tk1;
}

decimal Tk2(decimal t, decimal t5, decimal tl) {
	decimal tk2;
	if (t>=1.) {
		tk2=1-tl;
	} else if (t<=0.) {
		tk2=tl;
	} else if (tl>=1.) {
		tk2=1-t;
	} else if (tl<=0.) {
		tk2=t;
	} else {
		tk2=t + tl - 2*t*tl - 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl);
	}
	return tk2;
}

decimal Ck1(decimal t, decimal t5, decimal tl) {
	decimal ck1;
	if (t>=1.) {
		ck1=2.*tl-1.;
	} else if (t<=0.) {
		ck1=1.-2.*tl;
	} else if (tl>=1.) {
		ck1=2.*t-1.;
	} else if (tl<=0.) {
		ck1=1.-2.*t;
	} else {
		ck1=1.-2.*(t + tl - 2*t*tl + 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl));
	}
	return ck1;
}

decimal Ck2(decimal t, decimal t5, decimal tl) {
	decimal ck2;
	if (t>=1.) {
		ck2=2.*tl-1.;
	} else if (t<=0.) {
		ck2=1.-2.*tl;
	} else if (tl>=1.) {
		ck2=2.*t-1.;
	} else if (tl<=0.) {
		ck2=1.-2.*t;
	} else {
		ck2=1.-2.*(t + tl - 2*t*tl - 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl));
	}
	return ck2;
}

decimal Cl(decimal tl) {
	return 1.-2.*tl;
}

decimal Ck(decimal tk) {
	return 1.-2.*tk;
}

//get FXi from FX to hide structures from users
double FA1(decimal v,decimal tl,decimal y,decimal t5,decimal u,decimal b) {
	return FA(v,tl,y,Tk1(T(u,v),t5,tl),u,b);
}

double FA2(decimal v,decimal tl,decimal y,decimal t5,decimal u,decimal b) {
	return FA(v,tl,y,Tk2(T(u,v),t5,tl),u,b);
}

double FB1(decimal v,decimal tl,decimal y,decimal t5,decimal u,decimal b) {
	return FB(v,tl,y,Tk1(T(u,v),t5,tl),u,b);
}

double FB2(decimal v,decimal tl,decimal y,decimal t5,decimal u,decimal b) {
	return FB(v,tl,y,Tk2(T(u,v),t5,tl),u,b);
}

//get exponent M from variable m to hide structures from users
double M(double m) {
	double mm=2/m;
	if (isnan(mm) || isinf(mm) || mm==0. || mm<=0.) {
		return 2.;
	} else {
		return 2/m;
	}	
}

int D2IntCA(const int* ndim, const double x[], const int* ncomp, double F[],void *userdata) {
//CA structure
	F[0]=CAI2E0(x[0],x[1]);
	F[1]=CAI2EM1(x[0],x[1]);
	F[2]=CAI2EM2(x[0],x[1]);
	F[3]=CAI2EM3(x[0],x[1]);
	F[4]=CAI2EM4(x[0],x[1]);
//PNF structure
	F[5]=0;
	F[6]=0;
	F[7]=0;
	F[8]=0;
	return 1;
}

int D3IntCA(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
//CA structure
	F[0]=CAI3E0y(x[0],x[1],x[2])+CAI3E0t(x[0],x[1],x[2])+CAI3E0u(x[0],x[1],x[2])+CAI3E0b(x[0],x[1],x[2]);
	F[1]=CAI3EM1y(x[0],x[1],x[2])+CAI3EM1t(x[0],x[1],x[2])+CAI3EM1u(x[0],x[1],x[2])+CAI3EM1b(x[0],x[1],x[2]);
	F[2]=CAI3EM2y(x[0],x[1],x[2])+CAI3EM2t(x[0],x[1],x[2])+CAI3EM2u(x[0],x[1],x[2])+CAI3EM2b(x[0],x[1],x[2]);
	F[3]=CAI3EM3y(x[0],x[1],x[2])+CAI3EM3t(x[0],x[1],x[2])+CAI3EM3u(x[0],x[1],x[2])+CAI3EM3b(x[0],x[1],x[2]);
	F[4]=0;
//PNF structure
	F[5]=PNfI3E0(x[0],x[1],x[2]);
	F[6]=PNfI3EM1(x[0],x[1],x[2]);
	F[7]=PNfI3EM2(x[0],x[1],x[2]);
	F[8]=PNfI3EM3(x[0],x[1],x[2]);
		return 1;
}

int D3IntNF(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
	F[0]=NfI3E0(x[0],x[1],x[2]);
	F[1]=NfI3EM1(x[0],x[1],x[2]);
	F[2]=NfI3EM2(x[0],x[1],x[2]);
	F[3]=NfI3EM3(x[0],x[1],x[2]);
	return 1;
}

int D4IntCA(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
//CA structure
	F[0]=CAI4E0yt(x[0],x[1],x[2],x[3])+CAI4E0yu(x[0],x[1],x[2],x[3])+CAI4E0yb(x[0],x[1],x[2],x[3])+CAI4E0tu(x[0],x[1],x[2],x[3])+CAI4E0tb(x[0],x[1],x[2],x[3])+CAI4E0ub(x[0],x[1],x[2],x[3]);
	F[1]=CAI4EM1yt(x[0],x[1],x[2],x[3])+CAI4EM1yu(x[0],x[1],x[2],x[3])+CAI4EM1yb(x[0],x[1],x[2],x[3])+CAI4EM1tu(x[0],x[1],x[2],x[3])+CAI4EM1tb(x[0],x[1],x[2],x[3])+CAI4EM1ub(x[0],x[1],x[2],x[3]);
	F[2]=CAI4EM2yt(x[0],x[1],x[2],x[3])+CAI4EM2yu(x[0],x[1],x[2],x[3])+CAI4EM2yb(x[0],x[1],x[2],x[3])+CAI4EM2tu(x[0],x[1],x[2],x[3])+CAI4EM2tb(x[0],x[1],x[2],x[3])+CAI4EM2ub(x[0],x[1],x[2],x[3]);
	F[3]=0;
	F[4]=0;
//PNF structure
	F[5]=PNfI4E0y(x[0],x[1],x[2],x[3])+PNfI4E0t(x[0],x[1],x[2],x[3])+PNfI4E0u(x[0],x[1],x[2],x[3]);
	F[6]=PNfI4EM1y(x[0],x[1],x[2],x[3])+PNfI4EM1t(x[0],x[1],x[2],x[3])+PNfI4EM1u(x[0],x[1],x[2],x[3]);
	F[7]=PNfI4EM2y(x[0],x[1],x[2],x[3])+PNfI4EM2t(x[0],x[1],x[2],x[3])+PNfI4EM2u(x[0],x[1],x[2],x[3]);
	F[8]=0;
	return 1;
}

int D4IntNF(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
	F[0]=NfI4E0y(x[0],x[1],x[2],x[3])+NfI4E0t(x[0],x[1],x[2],x[3])+NfI4E0u(x[0],x[1],x[2],x[3]);
	F[1]=NfI4EM1y(x[0],x[1],x[2],x[3])+NfI4EM1t(x[0],x[1],x[2],x[3])+NfI4EM1u(x[0],x[1],x[2],x[3]);
	F[2]=NfI4EM2y(x[0],x[1],x[2],x[3])+NfI4EM2t(x[0],x[1],x[2],x[3])+NfI4EM2u(x[0],x[1],x[2],x[3]);
	F[3]=0;
	return 1;
}

int D5IntCA(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
//CA structure
	F[0]=CAI5E0y(x[0],x[1],x[2],x[3],x[4])+CAI5E0t(x[0],x[1],x[2],x[3],x[4])+CAI5E0u(x[0],x[1],x[2],x[3],x[4])+CAI5E0b(x[0],x[1],x[2],x[3],x[4]);
	F[1]=CAI5EM1y(x[0],x[1],x[2],x[3],x[4])+CAI5EM1t(x[0],x[1],x[2],x[3],x[4])+CAI5EM1u(x[0],x[1],x[2],x[3],x[4])+CAI5EM1b(x[0],x[1],x[2],x[3],x[4]);
	F[2]=0;
	F[3]=0;
	F[4]=0;
//PNF structure
	F[5]=PNfI5E0y(x[0],x[1],x[2],x[3],x[4])+PNfI5E0t(x[0],x[1],x[2],x[3],x[4])+PNfI5E0u(x[0],x[1],x[2],x[3],x[4]);
	F[6]=PNfI5EM1y(x[0],x[1],x[2],x[3],x[4])+PNfI5EM1t(x[0],x[1],x[2],x[3],x[4])+PNfI5EM1u(x[0],x[1],x[2],x[3],x[4]);
	F[7]=0;
	F[8]=0;
	return 1;
}

int D5IntNF(const int* ndim, const double x[],const int* ncomp,double F[],void* userdata) {
	F[0]=NfI5E0y(x[0],x[1],x[2],x[3],x[4])+NfI5E0t(x[0],x[1],x[2],x[3],x[4])+NfI5E0u(x[0],x[1],x[2],x[3],x[4]);
	F[1]=NfI5EM1y(x[0],x[1],x[2],x[3],x[4])+NfI5EM1t(x[0],x[1],x[2],x[3],x[4])+NfI5EM1u(x[0],x[1],x[2],x[3],x[4]);
	F[2]=0;
	F[3]=0;
	return 1;
}

int D2Int1P(const int* ndim, const double x[],const int* ncomp,double F[],void* userdata) {
	F[0]=IOLE3(x[0],x[1]);
	F[1]=IOLE2(x[0],x[1]);
	F[2]=IOLE1(x[0],x[1]);
	F[3]=IOLE0(x[0],x[1]);
	F[4]=IOLEM1(x[0],x[1]);
	F[5]=IRVE3(x[0],x[1]);
	F[6]=IRVE2(x[0],x[1]);
	F[7]=IRVE1(x[0],x[1]);
	F[8]=IRVE0(x[0],x[1]);
	F[9]=IRVEM1(x[0],x[1]);
	return 1;
}
