/*
 * This file contains the difference structures arising from plus distributions tying together the Measurement functions FXi, plus distribution denominators 
 * and contributing remainder/numerator functions XNfij for the 4D integration at all eps-orders and the alpha^0 order of the nf structure. 
 * Example: NfI4EM1s ('Nf' structure, 'I'ntegration over '4' dimensions, at 'E'ps to the 'M'inus '1'st and alpha^0 order, fourth integration over 's'5) 
 * contains logs of FXi functions evaluated at z=0,x=0, i.e. with the deltas from all +-distributions except the t5/s5 one, multiplied 
 * with the correct XNf functions to arrive at eps^-1 alpha^0. 
 */


#include <4IntNF.h>
#include <Auxiliaries.h>
#include <cmath>
#include <XFactorNf.h>
#include <Input_Common.h>

/*
 * the leading (eps^-2) terms vanish, since if x is the additional
 * integration the leading eps-term of this plus-dist vanishes,
 * and if x is delta'ed, it contributes a 1/alpha
 */

//fourth integration is over x

double NfI4EM1x(double w,double s6,double x,double c) {
	return ((-log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA1(v(w),tl(s6),y(x),0,0,b(c))) - log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),y(x),0,0,b(c))) - log(FB1(v(w),tl(s6),0,0,0,b(c))) +
		       log(FB1(v(w),tl(s6),y(x),0,0,b(c))) - log(FB2(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),y(x),0,0,b(c))))*pow(x,-1)*XNf00(w,s6,0,0,c))/2.;
}

double NfI4E0x(double w,double s6,double x,double c) {
	return (pow(x,-1)*(-2*(pow(log(FA1(v(w),tl(s6),0,0,0,b(c))),2) - pow(log(FA1(v(w),tl(s6),y(x),0,0,b(c))),2) + pow(log(FA2(v(w),tl(s6),0,0,0,b(c))),2) - pow(log(FA2(v(w),tl(s6),y(x),0,0,b(c))),2) +
	          pow(log(FB1(v(w),tl(s6),0,0,0,b(c))),2) - pow(log(FB1(v(w),tl(s6),y(x),0,0,b(c))),2) + pow(log(FB2(v(w),tl(s6),0,0,0,b(c))),2) - pow(log(FB2(v(w),tl(s6),y(x),0,0,b(c))),2))*
	        XNf00(w,s6,0,0,c) + (-log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA1(v(w),tl(s6),y(x),0,0,b(c))) - log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),y(x),0,0,b(c))) -
	          log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),y(x),0,0,b(c))) - log(FB2(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),y(x),0,0,b(c))))*XNf01(w,s6,0,0,c)))/2.;
}

//fourth integration is over s5

double NfI4EM1s(double w,double s6,double s5,double c) {
	return (pow(m,-1)*pow(s5,-1)*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf00(w,s6,0,0,c) -
		       (log(FA1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FA2(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB2(v(w),tl(s6),0,t5(s5),0,b(c))))*XNf00(w,s6,s5,0,c) +
		       2*(XNf10(w,s6,0,0,c) - XNf10(w,s6,s5,0,c))))/2.;
}

double NfI4E0s(double w,double s6,double s5,double c) {
	return (pow(m,-1)*pow(s5,-1)*(4*(pow(log(FA1(v(w),tl(s6),0,0,0,b(c))),2) + pow(log(FA2(v(w),tl(s6),0,0,0,b(c))),2) + pow(log(FB1(v(w),tl(s6),0,0,0,b(c))),2) +
	          pow(log(FB2(v(w),tl(s6),0,0,0,b(c))),2))*XNf00(w,s6,0,0,c) - 4*(pow(log(FA1(v(w),tl(s6),0,t5(s5),0,b(c))),2) + pow(log(FA2(v(w),tl(s6),0,t5(s5),0,b(c))),2) +
	          pow(log(FB1(v(w),tl(s6),0,t5(s5),0,b(c))),2) + pow(log(FB2(v(w),tl(s6),0,t5(s5),0,b(c))),2))*XNf00(w,s6,s5,0,c) +
	       (log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf01(w,s6,0,0,c) -
	       (log(FA1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FA2(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB2(v(w),tl(s6),0,t5(s5),0,b(c))))*XNf01(w,s6,s5,0,c) -
	       2*log(s5)*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf00(w,s6,0,0,c) -
	          (log(FA1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FA2(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB2(v(w),tl(s6),0,t5(s5),0,b(c))))*XNf00(w,s6,s5,0,c) +
	          2*(XNf10(w,s6,0,0,c) - XNf10(w,s6,s5,0,c))) + 2*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) +
	             log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf10(w,s6,0,0,c) + XNf11(w,s6,0,0,c)) -
	       2*((log(FA1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FA2(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB1(v(w),tl(s6),0,t5(s5),0,b(c))) + log(FB2(v(w),tl(s6),0,t5(s5),0,b(c))))*XNf10(w,s6,s5,0,c) +
	          XNf11(w,s6,s5,0,c))))/2.;
}

//fourth integration is over z

double NfI4EM1z(double w,double s6,double z,double c) {
	return pow(m,-1)*pow(z,-1)*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf00(w,s6,0,0,c) -
		     (log(FA1(v(w),tl(s6),0,0,u(z),b(c))) + log(FA2(v(w),tl(s6),0,0,u(z),b(c))) + log(FB1(v(w),tl(s6),0,0,u(z),b(c))) + log(FB2(v(w),tl(s6),0,0,u(z),b(c))))*XNf00(w,s6,0,z,c) +
		     2*(XNf10(w,s6,0,0,c) - XNf10(w,s6,0,z,c)));
}

double NfI4E0z(double w,double s6,double z,double c) {
	return pow(m,-1)*pow(z,-1)*(4*(pow(log(FA1(v(w),tl(s6),0,0,0,b(c))),2) + pow(log(FA2(v(w),tl(s6),0,0,0,b(c))),2) + pow(log(FB1(v(w),tl(s6),0,0,0,b(c))),2) + pow(log(FB2(v(w),tl(s6),0,0,0,b(c))),2))*
		      XNf00(w,s6,0,0,c) - 4*(pow(log(FA1(v(w),tl(s6),0,0,u(z),b(c))),2) + pow(log(FA2(v(w),tl(s6),0,0,u(z),b(c))),2) + pow(log(FB1(v(w),tl(s6),0,0,u(z),b(c))),2) +
		        pow(log(FB2(v(w),tl(s6),0,0,u(z),b(c))),2))*XNf00(w,s6,0,z,c) + (log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) +
		        log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf01(w,s6,0,0,c) - (log(FA1(v(w),tl(s6),0,0,u(z),b(c))) + log(FA2(v(w),tl(s6),0,0,u(z),b(c))) + log(FB1(v(w),tl(s6),0,0,u(z),b(c))) +
		        log(FB2(v(w),tl(s6),0,0,u(z),b(c))))*XNf01(w,s6,0,z,c) - 4*log(z)*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) +
		           log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf00(w,s6,0,0,c) - (log(FA1(v(w),tl(s6),0,0,u(z),b(c))) + log(FA2(v(w),tl(s6),0,0,u(z),b(c))) + log(FB1(v(w),tl(s6),0,0,u(z),b(c))) +
		           log(FB2(v(w),tl(s6),0,0,u(z),b(c))))*XNf00(w,s6,0,z,c) + 2*(XNf10(w,s6,0,0,c) - XNf10(w,s6,0,z,c))) +
		     2*((log(FA1(v(w),tl(s6),0,0,0,b(c))) + log(FA2(v(w),tl(s6),0,0,0,b(c))) + log(FB1(v(w),tl(s6),0,0,0,b(c))) + log(FB2(v(w),tl(s6),0,0,0,b(c))))*XNf10(w,s6,0,0,c) + XNf11(w,s6,0,0,c)) -
		     2*((log(FA1(v(w),tl(s6),0,0,u(z),b(c))) + log(FA2(v(w),tl(s6),0,0,u(z),b(c))) + log(FB1(v(w),tl(s6),0,0,u(z),b(c))) + log(FB2(v(w),tl(s6),0,0,u(z),b(c))))*XNf10(w,s6,0,z,c) +
		        XNf11(w,s6,0,z,c)));
}
