/* This function is the main body of the SoftSERVE project for the 1particle executable, i.e. the 1-loop bare soft function
 * and the 2-loop interference diagram. Due to the structure of the C language the program essentially creates child processes for each 
 * numerical integration (there's one for each number of dimensions in the integration domain), writes the integrator output to a log file
 * and collects the results of the various integrations. At the end of each integration the numerical results are combined with constant 
 * prefactors to get the coefficients of the bare soft function once the required results are available for each coefficient. The 
 * flow is commented along the flow.
 */ 

#include <Auxiliaries.h>
#include <Prefactors.h>
#include <cuba.h>
#include <IntOL.h>
#include <IntRV.h>
#include <Input_Common.h>

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

#include <stdlib.h>
#include <sys/types.h>
#include <algorithm>
#include <cstdlib>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/wait.h>

using namespace std;


	// contains the results of the numerical integration for OL and RV to (alpha,eps) orders {12,11,10,03,02,01,00,-14,-13,-12,-11,-10|03,02,01,00,-14,-13,-12,-11,-10}, where part after | is RV
	double res1P[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the error estimates for OL according to the scheme outlined above
	double err1P[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the probabilities that the error estimate is incorrect
	double prob1P[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};


	// collected result, error estimate and error probability for all regulator-orders
	double result[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double error[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double relprob[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	//variable containing the number of regions into which the integrator splits the integrand, irrelevant for us
	int nregions=0;

	/* The following variables are only necessary because of a specific problem: The Cuba integrators print their logs
	 * on stdout using puts() and fflush(). I want them to write to a file, but as I don't want to modify the Cuba
	 * sources, I have to work around that. There is a possibility of rerouting stdout to a file using freopen,
	 * but this is irreversible. Once stdout is rerouted, I can't send it back to the console. Solution: let Cuba
	 * run inside a child process that receives a copy of all file descriptors of the parent at its fork. If I
	 * freopen stdout in the child, let it terminate and then continue in the parent, the parent's stdout is still
	 * unaffected. Problem with that:
	 * Child and parent have different memory address spaces, each gets a copy of all variables and works only with
	 * those. So we need to introduce shared memory for the variables that the child modifies and the parent needs to
	 * read, i.e. the result, error estimate and error probability. */

	//process ids for the various child processes
	pid_t child;

	/*
	 * Shm always denotes something shared
	 */

	//memory ids for shared memory between parent and child
	int ShmResId,ShmErrId,ShmProbId,ShmNevId,ShmTerId;
	//pointers to shared memory for results, error estimates and error probabilities (these are arrays of type double)
	double *ShmRes, *ShmErr, *ShmProb;
	//pointer to shared process variable neval (# function evaluations), determined by child, but required by parent
	long long int *ShmNev;
	//pointer to shared process variable terminate, which is determined by child, but required by parent
	int *ShmTer;

	//file stream for output
	ofstream output;
	string outputname;

	 /*     ____             _
	  *    | __ )  ___  __ _(_)_ __    _ __  _ __ ___   __ _ _ __ __ _ _ __ ___
	  *    |  _ \ / _ \/ _` | | '_ \  | '_ \| '__/ _ \ / _` | '__/ _` | '_ ` _ \
	  *    | |_) |  __/ (_| | | | | | | |_) | | | (_) | (_| | | | (_| | | | | | |
	  *    |____/ \___|\__, |_|_| |_| | .__/|_|  \___/ \__, |_|  \__,_|_| |_| |_|
	  *                |___/          |_|              |___/
	  */

int main() {
	//open output file
	outputname="Result "+observable+" 1P.txt";
	output.open(outputname.c_str());

	//Info header
	cout << "Observable name: " << observable << "." << endl;
	cout << "Starting calculation for 1P results." << endl;

	/* Request shared memory for result ("Res", double type array of length ncomp1P),
	 * error estimate ("Err", double array of length ncomp1P),
	 * probability of wrong error ("Prob", double type array of length ncomp1P),
	 * neval variable ("Nev", long long int),
	 * termination code ("Ter", int);
	 *
	 * shmget returns the Id of the memory block (int type) if successful, -1 if it fails */
	ShmResId=shmget(IPC_PRIVATE, ncomp1P*sizeof(double), 0777|IPC_CREAT);
	ShmErrId=shmget(IPC_PRIVATE, ncomp1P*sizeof(double), 0777|IPC_CREAT);
	ShmProbId=shmget(IPC_PRIVATE, ncomp1P*sizeof(double), 0777|IPC_CREAT);
	ShmNevId=shmget(IPC_PRIVATE, sizeof(long long int), 0777|IPC_CREAT);
	ShmTerId=shmget(IPC_PRIVATE, sizeof(int), 0777|IPC_CREAT);

	//check if shared memory request was successful
	if(ShmResId<0 || ShmErrId <0 || ShmProbId<0 || ShmNevId<0 || ShmTerId<0) {
		cout << "Shared memory request failed." << endl;
		exit(EXIT_FAILURE);
	}

	/* Attach memory to parent's address space
	 *
	 * This construction is necessary, as shmat returns a pointer of type void*, which should be typecast
	 * to whatever we need (double* for ShmRes, ShmErr, ShmProb; long long int* for ShmNev; and int* for ShmTer).
	 * However, if shmat fails, it returns a char* -1, so to check I assign this return value to a void* pointer
	 * called Attx, check if it evaluates to -1 if typecast as char*, and if it doesn't (i.e. if the attachment
	 * was successful), I typecast Attx to double*, long long int*, or int*, and assign it to ShmRes, etc.	*/
	void *Att1,*Att2,*Att3,*Att4,*Att5;
	Att1 = shmat(ShmResId, NULL, 0);
	Att2 = shmat(ShmErrId, NULL, 0);
	Att3 = shmat(ShmProbId, NULL, 0);
	Att4 = shmat(ShmNevId, NULL, 0);
	Att5 = shmat(ShmTerId, NULL, 0);
	if((char*)Att1==(char*)-1 || (char*)Att2==(char*)-1 || (char*)Att3==(char*)-1 || (char*)Att4==(char*)-1 || (char*)Att5==(char*)-1) {
		cout << "Shared memory attachment failed" << endl;
		exit(EXIT_FAILURE);
	} else {
		ShmRes = (double*) Att1;
		ShmErr = (double*) Att2;
		ShmProb = (double*) Att3;
		ShmNev = (long long int*) Att4;
		ShmTer = (int*) Att5;
	}

	//start integration
	cout << "Numerical integration in progress..." << endl;

	//create child process and check if this was successful
	child=fork();
	if(child <0) { ///
		cout << "Fork for numerical integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncomp1P; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D2I1P= &D2Int1P;

		//reroute output of llDivonne to a text file
		if(freopen("1-Particle.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim2, ncomp1P, D2I1P,0, 1, epsrel2, epsabs2, flags, seed, 0, maxeval2, key12, key22, key32, maxpass2, border, maxchisq, mindev, 0, ndim2, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncomp1P),res1P);
	copy(ShmErr,(ShmErr+ncomp1P),err1P);
	copy(ShmProb,(ShmProb+ncomp1P),prob1P);

	//first integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " Numerical integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "Numerical integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}
	
	//compute results using prefactors for one loop (OL), assuming gaussian error combination and independent incorrect-error-estimate estimates
	//alpha^-1
	result[0]= PreOLA0EM1*res1P[11];
	error[0]=PreOLA0EM1*err1P[11];
	relprob[0]=prob1P[11];

	result[1]= PreOLA0EM1*res1P[10] + PreOLA0E0*res1P[11];
	error[1]=pow(pow(PreOLA0EM1,2)*pow(err1P[10],2) + pow(PreOLA0E0,2)*pow(err1P[11],2),0.5);
	relprob[1]=1 - (1 - prob1P[10])*(1 - prob1P[11]);

	result[2]=PreOLA0EM1*res1P[9] + PreOLA0E0*res1P[10] + PreOLA0E1*res1P[11];
	error[2]=pow(pow(PreOLA0EM1,2)*pow(err1P[9],2) + pow(PreOLA0E0,2)*pow(err1P[10],2) + pow(PreOLA0E1,2)*pow(err1P[11],2),0.5);
	relprob[2]=1 - (1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[3]=PreOLA0EM1*res1P[8] + PreOLA0E0*res1P[9] + PreOLA0E1*res1P[10] + PreOLA0E2*res1P[11];
	error[3]=pow(pow(PreOLA0EM1,2)*pow(err1P[8],2) + pow(PreOLA0E0,2)*pow(err1P[9],2) + pow(PreOLA0E1,2)*pow(err1P[10],2) + pow(PreOLA0E2,2)*pow(err1P[11],2),0.5);
	relprob[3]=1 - (1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[4]= PreOLA0EM1*res1P[7] + PreOLA0E0*res1P[8] + PreOLA0E1*res1P[9] + PreOLA0E2*res1P[10] + PreOLA0E3*res1P[11];
	error[4]= pow(pow(PreOLA0EM1,2)*pow(err1P[7],2) + pow(PreOLA0E0,2)*pow(err1P[8],2) + pow(PreOLA0E1,2)*pow(err1P[9],2) + pow(PreOLA0E2,2)*pow(err1P[10],2) + pow(PreOLA0E3,2)*pow(err1P[11],2),0.5);
	relprob[4]=1 - (1 - prob1P[7])*(1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);
	
	//alpha^0
	result[5]= PreOLA1EM2*res1P[11];
	error[5]= PreOLA1EM2*err1P[11];
	relprob[5]=prob1P[11];

	result[6]= PreOLA0EM1*res1P[6] + PreOLA1EM2*res1P[10] + PreOLA1EM1*res1P[11];
	error[6]= pow(pow(PreOLA0EM1,2)*pow(err1P[6],2) + pow(PreOLA1EM2,2)*pow(err1P[10],2) + pow(PreOLA1EM1,2)*pow(err1P[11],2),0.5);
	relprob[6]=1 - (1 - prob1P[6])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[7]= PreOLA0EM1*res1P[5] + PreOLA0E0*res1P[6] + PreOLA1EM2*res1P[9] + PreOLA1EM1*res1P[10] + PreOLA1E0*res1P[11];
	error[7]= pow(pow(PreOLA0EM1,2)*pow(err1P[5],2) + pow(PreOLA0E0,2)*pow(err1P[6],2) + pow(PreOLA1EM2,2)*pow(err1P[9],2) + pow(PreOLA1EM1,2)*pow(err1P[10],2) + pow(PreOLA1E0,2)*pow(err1P[11],2),0.5);
	relprob[7]=1 - (1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[8]= PreOLA0EM1*res1P[4] + PreOLA0E0*res1P[5] + PreOLA0E1*res1P[6] + PreOLA1EM2*res1P[8] + PreOLA1EM1*res1P[9] + PreOLA1E0*res1P[10] + PreOLA1E1*res1P[11];
	error[8]= pow(pow(PreOLA0EM1,2)*pow(err1P[4],2) + pow(PreOLA0E0,2)*pow(err1P[5],2) + pow(PreOLA0E1,2)*pow(err1P[6],2) + pow(PreOLA1EM2,2)*pow(err1P[8],2) + pow(PreOLA1EM1,2)*pow(err1P[9],2) + 
		    pow(PreOLA1E0,2)*pow(err1P[10],2) + pow(PreOLA1E1,2)*pow(err1P[11],2),0.5);
	relprob[8]=1 - (1 - prob1P[4])*(1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[9]= PreOLA0EM1*res1P[3] + PreOLA0E0*res1P[4] + PreOLA0E1*res1P[5] + PreOLA0E2*res1P[6] + PreOLA1EM2*res1P[7] + PreOLA1EM1*res1P[8] + PreOLA1E0*res1P[9] + PreOLA1E1*res1P[10] + PreOLA1E2*res1P[11];
	error[9]= pow(pow(PreOLA0EM1,2)*pow(err1P[3],2) + pow(PreOLA0E0,2)*pow(err1P[4],2) + pow(PreOLA0E1,2)*pow(err1P[5],2) + pow(PreOLA0E2,2)*pow(err1P[6],2) + pow(PreOLA1EM2,2)*pow(err1P[7],2) + 
		    pow(PreOLA1EM1,2)*pow(err1P[8],2) + pow(PreOLA1E0,2)*pow(err1P[9],2) + pow(PreOLA1E1,2)*pow(err1P[10],2) + pow(PreOLA1E2,2)*pow(err1P[11],2),0.5);
	relprob[9]=1 - (1 - prob1P[3])*(1 - prob1P[4])*(1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[7])*(1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	//alpha^1
	result[10]= PreOLA2EM3*res1P[11];
	error[10]= PreOLA2EM3*err1P[11];
	relprob[10]=prob1P[11];

	result[11]= PreOLA1EM2*res1P[6] + PreOLA2EM3*res1P[10] + PreOLA2EM2*res1P[11];
	error[11]= pow(pow(PreOLA1EM2,2)*pow(err1P[6],2) + pow(PreOLA2EM3,2)*pow(err1P[10],2) + pow(PreOLA2EM2,2)*pow(err1P[11],2),0.5);
	relprob[11]=1 - (1 - prob1P[6])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[12]= PreOLA0EM1*res1P[2] + PreOLA1EM2*res1P[5] + PreOLA1EM1*res1P[6] + PreOLA2EM3*res1P[9] + PreOLA2EM2*res1P[10] + PreOLA2EM1*res1P[11];
	error[12]= pow(pow(PreOLA0EM1,2)*pow(err1P[2],2) + pow(PreOLA1EM2,2)*pow(err1P[5],2) + pow(PreOLA1EM1,2)*pow(err1P[6],2) + pow(PreOLA2EM3,2)*pow(err1P[9],2) + pow(PreOLA2EM2,2)*pow(err1P[10],2) + 
		    pow(PreOLA2EM1,2)*pow(err1P[11],2),0.5);
	relprob[12]=1 - (1 - prob1P[2])*(1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[13]= PreOLA0EM1*res1P[1] + PreOLA0E0*res1P[2] + PreOLA1EM2*res1P[4] + PreOLA1EM1*res1P[5] + PreOLA1E0*res1P[6] + PreOLA2EM3*res1P[8] + PreOLA2EM2*res1P[9] + PreOLA2EM1*res1P[10] + PreOLA2E0*res1P[11];
	error[13]= pow(pow(PreOLA0EM1,2)*pow(err1P[1],2) + pow(PreOLA0E0,2)*pow(err1P[2],2) + pow(PreOLA1EM2,2)*pow(err1P[4],2) + pow(PreOLA1EM1,2)*pow(err1P[5],2) + pow(PreOLA1E0,2)*pow(err1P[6],2) + 
		    pow(PreOLA2EM3,2)*pow(err1P[8],2) + pow(PreOLA2EM2,2)*pow(err1P[9],2) + pow(PreOLA2EM1,2)*pow(err1P[10],2) + pow(PreOLA2E0,2)*pow(err1P[11],2),0.5);
	relprob[13]=1 - (1 - prob1P[1])*(1 - prob1P[2])*(1 - prob1P[4])*(1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11]);

	result[14]= PreOLA0E0*res1P[1] + PreOLA0E1*res1P[2] + PreOLA1EM2*res1P[3] + PreOLA1EM1*res1P[4] + PreOLA1E0*res1P[5] + PreOLA1E1*res1P[6] + PreOLA2EM3*res1P[7] + PreOLA2EM2*res1P[8] + PreOLA2EM1*res1P[9] + 
			   PreOLA2E0*res1P[10] + PreOLA2E1*res1P[11] + PreOLA0EM1*res1P[0];
	error[14]= pow(pow(PreOLA0E0,2)*pow(err1P[1],2) + pow(PreOLA0E1,2)*pow(err1P[2],2) + pow(PreOLA1EM2,2)*pow(err1P[3],2) + pow(PreOLA1EM1,2)*pow(err1P[4],2) + pow(PreOLA1E0,2)*pow(err1P[5],2) + 
		    pow(PreOLA1E1,2)*pow(err1P[6],2) + pow(PreOLA2EM3,2)*pow(err1P[7],2) + pow(PreOLA2EM2,2)*pow(err1P[8],2) + pow(PreOLA2EM1,2)*pow(err1P[9],2) + pow(PreOLA2E0,2)*pow(err1P[10],2) + 
		    pow(PreOLA2E1,2)*pow(err1P[11],2) + pow(PreOLA0EM1,2)*pow(err1P[0],2),0.5);
	relprob[14]=1 - (1 - prob1P[1])*(1 - prob1P[2])*(1 - prob1P[3])*(1 - prob1P[4])*(1 - prob1P[5])*(1 - prob1P[6])*(1 - prob1P[7])*(1 - prob1P[8])*(1 - prob1P[9])*(1 - prob1P[10])*(1 - prob1P[11])*(1 - prob1P[0]);
	
	//compute results for interference terms, assuming gaussian error combination and independent incorrect-error-estimate estimates
	//alpha^-1
	result[15]= PreRVA0EM3*res1P[20];
	error[15]= PreRVA0EM3*err1P[20];
	relprob[15]= prob1P[20];

	result[16]= PreRVA0EM3*res1P[19] + PreRVA0EM2*res1P[20];
	error[16]= pow(pow(PreRVA0EM3,2)*pow(err1P[19],2) + pow(PreRVA0EM2,2)*pow(err1P[20],2),0.5);
	relprob[16]= 1 - (1 - prob1P[19])*(1 - prob1P[20]);

	result[17]= PreRVA0EM3*res1P[18] + PreRVA0EM2*res1P[19] + PreRVA0EM1*res1P[20];
	error[17]= pow(pow(PreRVA0EM3,2)*pow(err1P[18],2) + pow(PreRVA0EM2,2)*pow(err1P[19],2) + pow(PreRVA0EM1,2)*pow(err1P[20],2),0.5);
	relprob[17]= 1 - (1 - prob1P[18])*(1 - prob1P[19])*(1 - prob1P[20]);

	result[18]= PreRVA0EM3*res1P[17] + PreRVA0EM2*res1P[18] + PreRVA0EM1*res1P[19] + PreRVA0E0*res1P[20];
	error[18]= pow(pow(PreRVA0EM3,2)*pow(err1P[17],2) + pow(PreRVA0EM2,2)*pow(err1P[18],2) + pow(PreRVA0EM1,2)*pow(err1P[19],2) + pow(PreRVA0E0,2)*pow(err1P[20],2),0.5);
	relprob[18]= 1 - (1 - prob1P[17])*(1 - prob1P[18])*(1 - prob1P[19])*(1 - prob1P[20]);

	//alpha^0
	result[19]= PreRVA1EM4*res1P[20];
	error[19]= PreRVA1EM4*err1P[20];
	relprob[19]= prob1P[20];

	result[20]= PreRVA0EM3*res1P[15] + PreRVA1EM4*res1P[19] + PreRVA1EM3*res1P[20];
	error[20]= pow(pow(PreRVA0EM3,2)*pow(err1P[15],2) + pow(PreRVA1EM4,2)*pow(err1P[19],2) + pow(PreRVA1EM3,2)*pow(err1P[20],2),0.5);
	relprob[20]=1 - (1 - prob1P[15])*(1 - prob1P[19])*(1 - prob1P[20]);

	result[21]= PreRVA0EM3*res1P[14] + PreRVA0EM2*res1P[15] + PreRVA1EM4*res1P[18] + PreRVA1EM3*res1P[19] + PreRVA1EM2*res1P[20];
	error[21]= pow(pow(PreRVA0EM3,2)*pow(err1P[14],2) + pow(PreRVA0EM2,2)*pow(err1P[15],2) + pow(PreRVA1EM4,2)*pow(err1P[18],2) + pow(PreRVA1EM3,2)*pow(err1P[19],2) + pow(PreRVA1EM2,2)*pow(err1P[20],2),0.5);
	relprob[21]=1 - (1 - prob1P[14])*(1 - prob1P[15])*(1 - prob1P[18])*(1 - prob1P[19])*(1 - prob1P[20]);

	result[22]= PreRVA0EM3*res1P[13] + PreRVA0EM2*res1P[14] + PreRVA0EM1*res1P[15] + PreRVA1EM4*res1P[17] + PreRVA1EM3*res1P[18] + PreRVA1EM2*res1P[19] + PreRVA1EM1*res1P[20];
	error[22]= pow(pow(PreRVA0EM3,2)*pow(err1P[13],2) + pow(PreRVA0EM2,2)*pow(err1P[14],2) + pow(PreRVA0EM1,2)*pow(err1P[15],2) + pow(PreRVA1EM4,2)*pow(err1P[17],2) + pow(PreRVA1EM3,2)*pow(err1P[18],2) + 
		    pow(PreRVA1EM2,2)*pow(err1P[19],2) + pow(PreRVA1EM1,2)*pow(err1P[20],2),0.5);
	relprob[22]=1 - (1 - prob1P[13])*(1 - prob1P[14])*(1 - prob1P[15])*(1 - prob1P[17])*(1 - prob1P[18])*(1 - prob1P[19])*(1 - prob1P[20]);

	result[23]= PreRVA0EM3*res1P[12] + PreRVA0EM2*res1P[13] + PreRVA0EM1*res1P[14] + PreRVA0E0*res1P[15] + PreRVA1EM4*res1P[16] + PreRVA1EM3*res1P[17] + PreRVA1EM2*res1P[18] + PreRVA1EM1*res1P[19] + PreRVA1E0*res1P[20];
	error[23]= pow(pow(PreRVA0EM3,2)*pow(err1P[12],2) + pow(PreRVA0EM2,2)*pow(err1P[13],2) + pow(PreRVA0EM1,2)*pow(err1P[14],2) + pow(PreRVA0E0,2)*pow(err1P[15],2) + pow(PreRVA1EM4,2)*pow(err1P[16],2) + 
		    pow(PreRVA1EM3,2)*pow(err1P[17],2) + pow(PreRVA1EM2,2)*pow(err1P[18],2) + pow(PreRVA1EM1,2)*pow(err1P[19],2) + pow(PreRVA1E0,2)*pow(err1P[20],2),0.5);
	relprob[23]=1 - (1 - prob1P[12])*(1 - prob1P[13])*(1 - prob1P[14])*(1 - prob1P[15])*(1 - prob1P[16])*(1 - prob1P[17])*(1 - prob1P[18])*(1 - prob1P[19])*(1 - prob1P[20]);

	//output results
	cout << "Results for One-Loop calculation: " << endl;
	output << observable << " is a SCET2 observable" << endl;
	cout << "One loop alpha^(-1) eps^(-1) coefficient: " << result[0] << " +/- " << error[0] << ", probability of incorrect error estimate: " << relprob[0] << endl;
	output << "One loop alpha^(-1) eps^(-1) coefficient: " << result[0] << " +/- " << error[0] << ", probability of incorrect error estimate: " << relprob[0] << endl;
	cout << "One loop alpha^(-1) eps^(0) coefficient: " << result[1] << " +/- " << error[1] << ", probability of incorrect error estimate: " << relprob[1] << endl;
	output << "One loop alpha^(-1) eps^(0) coefficient: " << result[1] << " +/- " << error[1] << ", probability of incorrect error estimate: " << relprob[1] << endl;
	cout << "One loop alpha^(-1) eps^(1) coefficient: " << result[2] << " +/- " << error[2] << ", probability of incorrect error estimate: " << relprob[2] << endl;
	output << "One loop alpha^(-1) eps^(1) coefficient: " << result[2] << " +/- " << error[2] << ", probability of incorrect error estimate: " << relprob[2] << endl;
	cout << "One loop alpha^(-1) eps^(2) coefficient: " << result[3] << " +/- " << error[3] << ", probability of incorrect error estimate: " << relprob[3] << endl;
	output << "One loop alpha^(-1) eps^(2) coefficient: " << result[3] << " +/- " << error[3] << ", probability of incorrect error estimate: " << relprob[3] << endl;
	cout << "One loop alpha^(-1) eps^(3) coefficient: " << result[4] << " +/- " << error[4] << ", probability of incorrect error estimate: " << relprob[4] << endl;
	output << "One loop alpha^(-1) eps^(3) coefficient: " << result[4] << " +/- " << error[4] << ", probability of incorrect error estimate: " << relprob[4] << endl;

	cout << "One loop alpha^(0) eps^(-2) coefficient: " << result[5] << " +/- " << error[5] << ", probability of incorrect error estimate: " << relprob[5] << endl;
	output << "One loop alpha^(0) eps^(-2) coefficient: " << result[5] << " +/- " << error[5] << ", probability of incorrect error estimate: " << relprob[5] << endl;
	cout << "One loop alpha^(0) eps^(-1) coefficient: " << result[6] << " +/- " << error[6] << ", probability of incorrect error estimate: " << relprob[6] << endl;
	output << "One loop alpha^(0) eps^(-1) coefficient: " << result[6] << " +/- " << error[6] << ", probability of incorrect error estimate: " << relprob[6] << endl;
	cout << "One loop alpha^(0) eps^(0) coefficient: " << result[7] << " +/- " << error[7] << ", probability of incorrect error estimate: " << relprob[7] << endl;
	output << "One loop alpha^(0) eps^(0) coefficient: " << result[7] << " +/- " << error[7] << ", probability of incorrect error estimate: " << relprob[7] << endl;
	cout << "One loop alpha^(0) eps^(1) coefficient: " << result[8] << " +/- " << error[8] << ", probability of incorrect error estimate: " << relprob[8] << endl;
	output << "One loop alpha^(0) eps^(1) coefficient: " << result[8] << " +/- " << error[8] << ", probability of incorrect error estimate: " << relprob[8] << endl;
	cout << "One loop alpha^(0) eps^(2) coefficient: " << result[9] << " +/- " << error[9] << ", probability of incorrect error estimate: " << relprob[9] << endl;
	output << "One loop alpha^(0) eps^(2) coefficient: " << result[9] << " +/- " << error[9] << ", probability of incorrect error estimate: " << relprob[9] << endl;

	cout << "One loop alpha^(1) eps^(-3) coefficient: " << result[10] << " +/- " << error[10] << ", probability of incorrect error estimate: " << relprob[10] << endl;
	output << "One loop alpha^(1) eps^(-3) coefficient: " << result[10] << " +/- " << error[10] << ", probability of incorrect error estimate: " << relprob[10] << endl;
	cout << "One loop alpha^(1) eps^(-2) coefficient: " << result[11] << " +/- " << error[11] << ", probability of incorrect error estimate: " << relprob[11] << endl;
	output << "One loop alpha^(1) eps^(-2) coefficient: " << result[11] << " +/- " << error[11] << ", probability of incorrect error estimate: " << relprob[11] << endl;
	cout << "One loop alpha^(1) eps^(-1) coefficient: " << result[12] << " +/- " << error[12] << ", probability of incorrect error estimate: " << relprob[12] << endl;
	output << "One loop alpha^(1) eps^(-1) coefficient: " << result[12] << " +/- " << error[12] << ", probability of incorrect error estimate: " << relprob[12] << endl;
	cout << "One loop alpha^(1) eps^(0) coefficient: " << result[13] << " +/- " << error[13] << ", probability of incorrect error estimate: " << relprob[13] << endl;
	output << "One loop alpha^(1) eps^(0) coefficient: " << result[13] << " +/- " << error[13] << ", probability of incorrect error estimate: " << relprob[13] << endl;
	cout << "One loop alpha^(1) eps^(1) coefficient: " << result[14] << " +/- " << error[14] << ", probability of incorrect error estimate: " << relprob[14] << endl;
	output << "One loop alpha^(1) eps^(1) coefficient: " << result[14] << " +/- " << error[14] << ", probability of incorrect error estimate: " << relprob[14] << endl;

	
	cout << "Results for interference diagram calculation: " << endl;
	cout << "RV alpha^(-1) eps^(-3) coefficient: " << result[15] << " +/- " << error[15] << ", probability of incorrect error estimate: " << relprob[15] << endl;
	output << "RV alpha^(-1) eps^(-3) coefficient: " << result[15] << " +/- " << error[15] << ", probability of incorrect error estimate: " << relprob[15] << endl;
	cout << "RV alpha^(-1) eps^(-2) coefficient: " << result[16] << " +/- " << error[16] << ", probability of incorrect error estimate: " << relprob[16] << endl;
	output << "RV alpha^(-1) eps^(-2) coefficient: " << result[16] << " +/- " << error[16] << ", probability of incorrect error estimate: " << relprob[16] << endl;
	cout << "RV alpha^(-1) eps^(-1) coefficient: " << result[17] << " +/- " << error[17] << ", probability of incorrect error estimate: " << relprob[17] << endl;
	output << "RV alpha^(-1) eps^(-1) coefficient: " << result[17] << " +/- " << error[17] << ", probability of incorrect error estimate: " << relprob[17] << endl;
	cout << "RV alpha^(-1) eps^(0) coefficient: " << result[18] << " +/- " << error[18] << ", probability of incorrect error estimate: " << relprob[18] << endl;
	output << "RV alpha^(-1) eps^(0) coefficient: " << result[18] << " +/- " << error[18] << ", probability of incorrect error estimate: " << relprob[18] << endl;

	cout << "RV alpha^(0) eps^(-4) coefficient: " << result[19] << " +/- " << error[19] << ", probability of incorrect error estimate: " << relprob[19] << endl;
	output << "RV alpha^(0) eps^(-4) coefficient: " << result[19] << " +/- " << error[19] << ", probability of incorrect error estimate: " << relprob[19] << endl;
	cout << "RV alpha^(0) eps^(-3) coefficient: " << result[20] << " +/- " << error[20] << ", probability of incorrect error estimate: " << relprob[20] << endl;
	output << "RV alpha^(0) eps^(-3) coefficient: " << result[20] << " +/- " << error[20] << ", probability of incorrect error estimate: " << relprob[20] << endl;
	cout << "RV alpha^(0) eps^(-2) coefficient: " << result[21] << " +/- " << error[21] << ", probability of incorrect error estimate: " << relprob[21] << endl;
	output << "RV alpha^(0) eps^(-2) coefficient: " << result[21] << " +/- " << error[21] << ", probability of incorrect error estimate: " << relprob[21] << endl;
	cout << "RV alpha^(0) eps^(-1) coefficient: " << result[22] << " +/- " << error[22] << ", probability of incorrect error estimate: " << relprob[22] << endl;
	output << "RV alpha^(0) eps^(-1) coefficient: " << result[22] << " +/- " << error[22] << ", probability of incorrect error estimate: " << relprob[22] << endl;
	cout << "RV alpha^(0) eps^(0) coefficient: " << result[23] << " +/- " << error[23] << ", probability of incorrect error estimate: " << relprob[23] << endl;
	output << "RV alpha^(0) eps^(0) coefficient: " << result[23] << " +/- " << error[23] << ", probability of incorrect error estimate: " << relprob[23] << endl;

	//detach shared memory
	shmdt(ShmRes);
	shmdt(ShmErr);
	shmdt(ShmProb);
	shmdt(ShmNev);
	shmdt(ShmTer);

	//delete/destroy shared memory
	shmctl(ShmResId, IPC_RMID, 0);
	shmctl(ShmErrId, IPC_RMID, 0);
	shmctl(ShmProbId, IPC_RMID, 0);
	shmctl(ShmNevId, IPC_RMID, 0);
	shmctl(ShmTerId, IPC_RMID, 0);
	return 0;

	//close output file
	output.close();
}
