/* This function is the main body of the SoftSERVE project for the CA structure executable of the 2-loop bare soft function
 * Due to the structure of the C language the program essentially creates child processes for each 
 * numerical integration (there's one for each number of dimensions in the integration domain), writes the integrator output to a log file
 * and collects the results of the various integrations. At the end of each integration the numerical results are combined with constant 
 * prefactors to get the coefficients of the bare soft function once the required results are available for each coefficient. The 
 * flow is commented along the flow.
 */ 

#include <Auxiliaries.h>
#include <2IntCA.h>
#include <3IntCA.h>
#include <4IntCA.h>
#include <5IntCA.h>
#include <3IntPNF.h>
#include <4IntPNF.h>
#include <5IntPNF.h>
#include <2IntCAAM1.h>
#include <3IntCAAM1.h>
#include <4IntCAAM1.h>
#include <5IntCAAM1.h>
#include <3IntPNFAM1.h>
#include <4IntPNFAM1.h>
#include <5IntPNFAM1.h>
#include <Prefactors.h>
#include <Input_Common.h>
#include <cuba.h>

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

#include <stdlib.h>
#include <sys/types.h>
#include <algorithm>
#include <cstdlib>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/wait.h>

using namespace std;

	/* contains the results of the n-dim integration for CA to (alpha, eps)
	 * orders ((0,0),(0,-1),(0,-2),(0,-3),(-1,1),(-1,0),(-1,-1),(-1,-2),(-1,-3),
	 * (0,0),(0,-1),(0,-2),(-1,1),(-1,0),(-1,-1),(-1,-2))
	 * where the second block starting with (0,0) is Pseudo-Nf */
	double resca2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double resca3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double resca4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double resca5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the error estimates for n-dim CA according to the scheme outlined above
	double errca2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double errca3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double errca4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double errca5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the probabilities that a certain n-dim integration's error estimate is incorrect
	double probca2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double probca3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double probca4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double probca5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	/* collected result, error estimate and error probability for a certain eps-order
	 * as soon as the necessary data is available. these variables are only buffers, they are reused
	 * for different eps-orders. The -al suffix denotes 1/alpha results */
	double result=0;
	double error=0;
	double relprob=0;
	double resultal=0;
	double erroral=0;
	double relprobal=0;

	//variable containing the number of regions into which the integrator splits the integrand, irrelevant for us
	int nregions=0;

	/* The following variables are only necessary because of a specific problem: The Cuba integrators print their logs
	 * on stdout using puts() and fflush(). I want them to write to a file, but as I don't want to modify the Cuba
	 * sources, I have to work around that. There is a possibility of rerouting stdout to a file using freopen,
	 * but this is irreversible. Once stdout is rerouted, I can't send it back to the console. Solution: let Cuba
	 * run inside a child process that receives a copy of all file descriptors of the parent at its fork. If I
	 * freopen stdout in the child, let it terminate and then continue in the parent, the parent's stdout is still
	 * unaffected. Problem with that:
	 * Child and parent have different memory address spaces, each gets a copy of all variables and works only with
	 * those. So we need to introduce shared memory for the variables that the child modifies and the parent needs to
	 * read, i.e. the result, error estimate and error probability. */

	//process ids for the various child processes
	pid_t child2, child3, child4, child5;

	/*
	 * Shm always denotes something shared
	 */

	//memory ids for shared memory between parent and child
	int ShmResId,ShmErrId,ShmProbId,ShmNevId,ShmTerId;
	//pointers to shared memory for results, error estimates and error probabilities (these are arrays of type double)
	double *ShmRes, *ShmErr, *ShmProb;
	//pointer to shared process variable neval (# function evaluations), determined by child, but required by parent
	long long int *ShmNev;
	//pointer to shared process variable terminate, which is determined by child, but required by parent
	int *ShmTer;

	//file stream for output
	ofstream output;
	string outputname;

	 /*     ____             _
	  *    | __ )  ___  __ _(_)_ __    _ __  _ __ ___   __ _ _ __ __ _ _ __ ___
	  *    |  _ \ / _ \/ _` | | '_ \  | '_ \| '__/ _ \ / _` | '__/ _` | '_ ` _ \
	  *    | |_) |  __/ (_| | | | | | | |_) | | | (_) | (_| | | | (_| | | | | | |
	  *    |____/ \___|\__, |_|_| |_| | .__/|_|  \___/ \__, |_|  \__,_|_| |_| |_|
	  *                |___/          |_|              |___/
	  */

int main() {
	//open output file
	outputname="Result "+observable+" Ca.txt";
	output.open(outputname.c_str());
	
	//Info header
	cout << "Observable name: " << observable << "." << endl;
	cout << "Starting calculation for CA structure." << endl;

	/* Request shared memory for result ("Res", double array of length ncompCA),
	 * error estimate ("Err", double array of length ncompCA),
	 * probability of wrong error ("Prob", double array of length ncompCA),
	 * neval variable ("Nev", long long int),
	 * termination code ("Ter", int);
	 *
	 * shmget returns the ID of the memory block (int type) if successful, -1 if it fails */
	ShmResId=shmget(IPC_PRIVATE, ncompCA*sizeof(double), 0777|IPC_CREAT);
	ShmErrId=shmget(IPC_PRIVATE, ncompCA*sizeof(double), 0777|IPC_CREAT);
	ShmProbId=shmget(IPC_PRIVATE, ncompCA*sizeof(double), 0777|IPC_CREAT);
	ShmNevId=shmget(IPC_PRIVATE, sizeof(long long int), 0777|IPC_CREAT);
	ShmTerId=shmget(IPC_PRIVATE, sizeof(int), 0777|IPC_CREAT);

	//check if shared memory request was successful
	if(ShmResId<0 || ShmErrId <0 || ShmProbId<0 || ShmNevId<0 || ShmTerId<0) {
		cout << "Shared memory request failed." << endl;
		exit(EXIT_FAILURE);
	}

	/* Attach memory to parent's address space
	 *
	 * This construction is necessary, as shmat returns a pointer of type void*, which should be typecast
	 * to whatever we need (double* for ShmRes, ShmErr, ShmProb; long long int* for ShmNev; and int* for ShmTer).
	 * However, if shmat fails, it returns a char* -1, so to check I assign this return value to a void* pointer
	 * called Attx, check if it evaluates to -1 if typecast as char*, and if it doesn't (i.e. if the attachment
	 * was successful), I typecast Attx to double*, long long int*, or int*, and assign it to ShmRes, etc.	*/
	void *Att1,*Att2,*Att3,*Att4,*Att5;
	Att1 = shmat(ShmResId, NULL, 0);
	Att2 = shmat(ShmErrId, NULL, 0);
	Att3 = shmat(ShmProbId, NULL, 0);
	Att4 = shmat(ShmNevId, NULL, 0);
	Att5 = shmat(ShmTerId, NULL, 0);
	if((char*)Att1==(char*)-1 || (char*)Att2==(char*)-1 || (char*)Att3==(char*)-1 || (char*)Att4==(char*)-1 || (char*)Att5==(char*)-1) {
		cout << "Shared memory attachment failed" << endl;
		exit(EXIT_FAILURE);
	} else {
		ShmRes = (double*) Att1;
		ShmErr = (double*) Att2;
		ShmProb = (double*) Att3;
		ShmNev = (long long int*) Att4;
		ShmTer = (int*) Att5;
	}

	//finally, start first integration
	cout << "2-dimensional integration in progress..." << endl;
	//create child process and check if this was successful
	child2=fork();
	if(child2 <0) { ///
		cout << "Fork for 2-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child2==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncompCA; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources (D2ICa = 'D'imension'2I'ntegrationfor'CA')
		integrand_t D2ICA= &D2IntCA;

		//reroute output of llDivonne to a text file
		if(freopen("2-dimensional CA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim2, ncompCA, D2ICA,0, 1, epsrel2, epsabs2, flags, seed, 0, maxeval2, key12, key22, key32, maxpass2, border, maxchisq, mindev, 0, ndim2, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before the parent does anything */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncompCA),resca2);
	copy(ShmErr,(ShmErr+ncompCA),errca2);
	copy(ShmProb,(ShmProb+ncompCA),probca2);

	//first integration finished
	if(*ShmTer<0) {
		cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
	} else if(*ShmTer>0) {
		cout << " 2-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
	} else if(*ShmTer==0){
		cout << "2-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
	} else {
		cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
	}

	//compute first result (2-dim integration alone is sufficient for div^(-4) coefficient), assuming gaussian error combination and independent incorrect-error-estimate estimates
	result=PreCA00*resca2[4] + PreCA1M1*resca2[12];
	resultal=PreCA00*resca2[12];
	error=pow(pow(PreCA00,2)*pow(errca2[4],2) + pow(PreCA1M1,2)*pow(errca2[12],2),0.5);
	erroral=PreCA00*errca2[12];
	relprob=1 - (1 - probca2[4])*(1 - probca2[12]);
	relprobal=probca2[12];

	//announce the result to the world at large
	cout << "Sufficient data for alpha^(0) eps^(-4) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-4) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-3) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-3) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;

	//start second integration
	cout << "3-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child3=fork();
	if(child3 <0) { ///
		cout << "Fork for 3-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child3==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncompCA; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D3ICA= &D3IntCA;

		//reroute output of llDivonne to a text file
		if(freopen("3-dimensional CA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim3, ncompCA, D3ICA,0, 1, epsrel3, epsabs3, flags, seed, 0, maxeval3, key13, key23, key33, maxpass3, border, maxchisq, mindev, 0, ndim3, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncompCA),resca3);
	copy(ShmErr,(ShmErr+ncompCA),errca3);
	copy(ShmProb,(ShmProb+ncompCA),probca3);

	//second integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 3-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "3-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}

	//compute second result (div^(-3) needs only 2-D and 3-D), assuming gaussian error combination and independent incorrect-error-estimate estimates
	result=PreCA00*resca2[3] + PreCA01*resca2[4] + PreCA1M1*resca2[11] + PreCA10*resca2[12] + PreCA00*resca3[3] + PreCA1M1*resca3[11] +
			 PrePNF1M1*resca3[16];
	error=pow(pow(PreCA00,2)*pow(errca2[3],2) + pow(PreCA01,2)*pow(errca2[4],2) + pow(PreCA1M1,2)*pow(errca2[11],2) +
		    pow(PreCA10,2)*pow(errca2[12],2) + pow(PreCA00,2)*pow(errca3[3],2) + pow(PreCA1M1,2)*pow(errca3[11],2) +
		    pow(PrePNF1M1,2)*pow(errca3[16],2),0.5);
	relprob=1 - (1 - probca2[3])*(1 - probca2[4])*(1 - probca2[11])*(1 - probca2[12])*(1 - probca3[3])*(1 - probca3[11])*(1 - probca3[16]);
	resultal=PreCA00*resca2[11] + PreCA01*resca2[12] + PreCA00*resca3[11] + PrePNF00*resca3[16];
	erroral=pow(pow(PreCA00,2)*pow(errca2[11],2) + pow(PreCA01,2)*pow(errca2[12],2) + pow(PreCA00,2)*pow(errca3[11],2) +
		    pow(PrePNF00,2)*pow(errca3[16],2),0.5);
	relprobal=1 - (1 - probca2[11])*(1 - probca2[12])*(1 - probca3[11])*(1 - probca3[16]);

	cout << "Sufficient data for alpha^(0) eps^(-3) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-3) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-2) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-2) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;

	//start third integration
	cout << "4-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child4=fork();
	if(child4 <0) { ///
		cout << "Fork for 4-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child4==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncompCA; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D4ICA= &D4IntCA;

		//reroute output of llDivonne to a text file
		if(freopen("4-dimensional CA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim4, ncompCA, D4ICA,0, 1, epsrel4, epsabs4, flags, seed, 0, maxeval4, key14, key24, key34, maxpass4, border, maxchisq, mindev, 0, ndim4, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);
		
		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncompCA),resca4);
	copy(ShmErr,(ShmErr+ncompCA),errca4);
	copy(ShmProb,(ShmProb+ncompCA),probca4);

	//third integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 4-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "4-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}
	//compute third result (div^(-2) needs 2-D, 3-D and 4-D), assuming gaussian error combination and independent incorrect-error-estimate estimates
	result=PreCA00*resca2[2] + PreCA01*resca2[3] + PreCA02*resca2[4] + PreCA00*resca3[2] + PreCA01*resca3[3] + PreCA00*resca4[2]
		 + PreCA1M1*resca2[10] + PreCA10*resca2[11] + PreCA11*resca2[12] + PreCA1M1*resca3[10] + PreCA10*resca3[11] + PreCA1M1*resca4[10]
		 + PrePNF00*resca3[7] + PrePNF00*resca4[7]
		 + PrePNF1M1*resca3[15] + PrePNF10*resca3[16] + PrePNF1M1*resca4[15];
	error=pow(pow(PreCA00,2)*pow(errca2[2],2) + pow(PreCA01,2)*pow(errca2[3],2) + pow(PreCA02,2)*pow(errca2[4],2) +
		    pow(PreCA1M1,2)*pow(errca2[10],2) + pow(PreCA10,2)*pow(errca2[11],2) + pow(PreCA11,2)*pow(errca2[12],2) +
		    pow(PreCA00,2)*pow(errca3[2],2) + pow(PreCA01,2)*pow(errca3[3],2) + pow(PrePNF00,2)*pow(errca3[7],2) +
		    pow(PreCA1M1,2)*pow(errca3[10],2) + pow(PreCA10,2)*pow(errca3[11],2) + pow(PrePNF1M1,2)*pow(errca3[15],2) +
		    pow(PrePNF10,2)*pow(errca3[16],2) + pow(PreCA00,2)*pow(errca4[2],2) + pow(PrePNF00,2)*pow(errca4[7],2) +
		    pow(PreCA1M1,2)*pow(errca4[10],2) + pow(PrePNF1M1,2)*pow(errca4[15],2),0.5);
	relprob=1 - (1 - probca2[2])*(1 - probca2[3])*(1 - probca2[4])*(1 - probca2[10])*(1 - probca2[11])*(1 - probca2[12])*(1 - probca3[2])*
			  (1 - probca3[3])*(1 - probca3[7])*(1 - probca3[10])*(1 - probca3[11])*(1 - probca3[15])*(1 - probca3[16])*(1 - probca4[2])*
			  (1 - probca4[7])*(1 - probca4[10])*(1 - probca4[15]);

	resultal=PreCA02*resca2[12] + PreCA01*resca2[11] + PreCA00*resca2[10] + PreCA01*resca3[11] + PreCA00*resca3[10] + PreCA00*resca4[10]
			+ PrePNF01*resca3[16] + PrePNF00*resca3[15] + PrePNF00*resca4[15];
	erroral=pow(pow(PreCA00,2)*pow(errca2[10],2) + pow(PreCA01,2)*pow(errca2[11],2) + pow(PreCA02,2)*pow(errca2[12],2) +
		    pow(PreCA00,2)*pow(errca3[10],2) + pow(PreCA01,2)*pow(errca3[11],2) + pow(PrePNF00,2)*pow(errca3[15],2) +
		    pow(PrePNF01,2)*pow(errca3[16],2) + pow(PreCA00,2)*pow(errca4[10],2) + pow(PrePNF00,2)*pow(errca4[15],2),0.5);
	relprobal=1 - (1 - probca2[10])*(1 - probca2[11])*(1 - probca2[12])*(1 - probca3[10])*(1 - probca3[11])*(1 - probca3[15])*
			  (1 - probca3[16])*(1 - probca4[10])*(1 - probca4[15]);

	cout << "Sufficient data for alpha^(0) eps^(-2) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-2) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-1) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-1) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;
	
	//start fourth integration
	cout << "5-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child5=fork();
	if(child5 <0) { ///
		cout << "Fork for 5-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child5==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncompCA; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D5ICA= &D5IntCA;

		//reroute output of llDivonne to a text file
		if(freopen("5-dimensional CA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim5, ncompCA, D5ICA,0, 1, epsrel5, epsabs5, flags, seed, 0, maxeval5, key15, key25, key35, maxpass5, border, maxchisq, mindev, 0, ndim5, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);
		
		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated well before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncompCA),resca5);
	copy(ShmErr,(ShmErr+ncompCA),errca5);
	copy(ShmProb,(ShmProb+ncompCA),probca5);

	//fourth integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 5-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "5-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}

	//compute last results (div^(-1) and div^0 need 2-D to 5-D, 6-D is not required for div^0, counterintuitively)
	//We assume gaussian error combination and independent incorrect-error-estimate estimates
	result=PreCA00*resca2[1] + PreCA01*resca2[2] + PreCA02*resca2[3] + PreCA03*resca2[4] + PreCA1M1*resca2[9] + PreCA10*resca2[10] +
			 PreCA11*resca2[11] + PreCA12*resca2[12] + PreCA00*resca3[1] + PreCA01*resca3[2] + PreCA02*resca3[3] + PrePNF00*resca3[6] +
			 PrePNF01*resca3[7] + PreCA1M1*resca3[9] + PreCA10*resca3[10] + PreCA11*resca3[11] + PrePNF1M1*resca3[14] +
			 PrePNF10*resca3[15] + PrePNF11*resca3[16] + PreCA00*resca4[1] + PreCA01*resca4[2] + PrePNF00*resca4[6] + PrePNF01*resca4[7] +
			 PreCA1M1*resca4[9] + PreCA10*resca4[10] + PrePNF1M1*resca4[14] + PrePNF10*resca4[15] + PreCA00*resca5[1] + PrePNF00*resca5[6] +
			 PreCA1M1*resca5[9] + PrePNF1M1*resca5[14];
	error=pow(pow(PreCA00,2)*pow(errca2[1],2) + pow(PreCA01,2)*pow(errca2[2],2) + pow(PreCA02,2)*pow(errca2[3],2) +
		    pow(PreCA03,2)*pow(errca2[4],2) + pow(PreCA1M1,2)*pow(errca2[9],2) + pow(PreCA10,2)*pow(errca2[10],2) +
		    pow(PreCA11,2)*pow(errca2[11],2) + pow(PreCA12,2)*pow(errca2[12],2) + pow(PreCA00,2)*pow(errca3[1],2) +
		    pow(PreCA01,2)*pow(errca3[2],2) + pow(PreCA02,2)*pow(errca3[3],2) + pow(PrePNF00,2)*pow(errca3[6],2) +
		    pow(PrePNF01,2)*pow(errca3[7],2) + pow(PreCA1M1,2)*pow(errca3[9],2) + pow(PreCA10,2)*pow(errca3[10],2) +
		    pow(PreCA11,2)*pow(errca3[11],2) + pow(PrePNF1M1,2)*pow(errca3[14],2) + pow(PrePNF10,2)*pow(errca3[15],2) +
		    pow(PrePNF11,2)*pow(errca3[16],2) + pow(PreCA00,2)*pow(errca4[1],2) + pow(PreCA01,2)*pow(errca4[2],2) +
		    pow(PrePNF00,2)*pow(errca4[6],2) + pow(PrePNF01,2)*pow(errca4[7],2) + pow(PreCA1M1,2)*pow(errca4[9],2) +
		    pow(PreCA10,2)*pow(errca4[10],2) + pow(PrePNF1M1,2)*pow(errca4[14],2) + pow(PrePNF10,2)*pow(errca4[15],2) +
		    pow(PreCA00,2)*pow(errca5[1],2) + pow(PrePNF00,2)*pow(errca5[6],2) + pow(PreCA1M1,2)*pow(errca5[9],2) +
		    pow(PrePNF1M1,2)*pow(errca5[14],2),0.5);
	relprob=1 - (1 - probca2[1])*(1 - probca2[2])*(1 - probca2[3])*(1 - probca2[4])*(1 - probca2[9])*(1 - probca2[10])*(1 - probca2[11])*
			  (1 - probca2[12])*(1 - probca3[1])*(1 - probca3[2])*(1 - probca3[3])*(1 - probca3[6])*(1 - probca3[7])*(1 - probca3[9])*
			  (1 - probca3[10])*(1 - probca3[11])*(1 - probca3[14])*(1 - probca3[15])*(1 - probca3[16])*(1 - probca4[1])*
			  (1 - probca4[2])*(1 - probca4[6])*(1 - probca4[7])*(1 - probca4[9])*(1 - probca4[10])*(1 - probca4[14])*(1 - probca4[15])*
			  (1 - probca5[1])*(1 - probca5[6])*(1 - probca5[9])*(1 - probca5[14]);
	resultal=PreCA00*resca2[9] + PreCA01*resca2[10] + PreCA02*resca2[11] + PreCA03*resca2[12] + PreCA00*resca3[9] + PreCA01*resca3[10] +
			 PreCA02*resca3[11] + PrePNF00*resca3[14] + PrePNF01*resca3[15] + PrePNF02*resca3[16] + PreCA00*resca4[9] + PreCA01*resca4[10] +
			 PrePNF00*resca4[14] + PrePNF01*resca4[15] + PreCA00*resca5[9] + PrePNF00*resca5[14];
	erroral=pow(pow(PreCA00,2)*pow(errca2[9],2) + pow(PreCA01,2)*pow(errca2[10],2) + pow(PreCA02,2)*pow(errca2[11],2) +
		    pow(PreCA03,2)*pow(errca2[12],2) + pow(PreCA00,2)*pow(errca3[9],2) + pow(PreCA01,2)*pow(errca3[10],2) +
		    pow(PreCA02,2)*pow(errca3[11],2) + pow(PrePNF00,2)*pow(errca3[14],2) + pow(PrePNF01,2)*pow(errca3[15],2) +
		    pow(PrePNF02,2)*pow(errca3[16],2) + pow(PreCA00,2)*pow(errca4[9],2) + pow(PreCA01,2)*pow(errca4[10],2) +
		    pow(PrePNF00,2)*pow(errca4[14],2) + pow(PrePNF01,2)*pow(errca4[15],2) + pow(PreCA00,2)*pow(errca5[9],2) +
		    pow(PrePNF00,2)*pow(errca5[14],2),0.5);
	relprobal=1 - (1 - probca2[9])*(1 - probca2[10])*(1 - probca2[11])*(1 - probca2[12])*(1 - probca3[9])*(1 - probca3[10])*(1 - probca3[11])*
			  (1 - probca3[14])*(1 - probca3[15])*(1 - probca3[16])*(1 - probca4[9])*(1 - probca4[10])*(1 - probca4[14])*(1 - probca4[15])*
			  (1 - probca5[9])*(1 - probca5[14]);

	cout << "Sufficient data for alpha^(0) eps^(-1) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-1) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(0) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(0) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;

	// final result for (0,0) order, assuming gaussian error combination and independent incorrect-error-estimate estimates
	result=PreCA00*resca2[0] + PreCA01*resca2[1] + PreCA02*resca2[2] + PreCA03*resca2[3] + PreCA04*resca2[4] + PreCA1M1*resca2[8] +
			 PreCA10*resca2[9] + PreCA11*resca2[10] + PreCA12*resca2[11] + PreCA13*resca2[12] + PreCA00*resca3[0] + PreCA01*resca3[1] +
			 PreCA02*resca3[2] + PreCA03*resca3[3] + PrePNF00*resca3[5] + PrePNF01*resca3[6] + PrePNF02*resca3[7] + PreCA1M1*resca3[8] +
			 PreCA10*resca3[9] + PreCA11*resca3[10] + PreCA12*resca3[11] + PrePNF1M1*resca3[13] + PrePNF10*resca3[14] +
			 PrePNF11*resca3[15] + PrePNF12*resca3[16] + PreCA00*resca4[0] + PreCA01*resca4[1] + PreCA02*resca4[2] + PrePNF00*resca4[5] +
			 PrePNF01*resca4[6] + PrePNF02*resca4[7] + PreCA1M1*resca4[8] + PreCA10*resca4[9] + PreCA11*resca4[10] + PrePNF1M1*resca4[13] +
			 PrePNF10*resca4[14] + PrePNF11*resca4[15] + PreCA00*resca5[0] + PreCA01*resca5[1] + PrePNF00*resca5[5] + PrePNF01*resca5[6] +
			 PreCA1M1*resca5[8] + PreCA10*resca5[9] + PrePNF1M1*resca5[13] + PrePNF10*resca5[14];
	error=pow(pow(PreCA00,2)*pow(errca2[0],2) + pow(PreCA01,2)*pow(errca2[1],2) + pow(PreCA02,2)*pow(errca2[2],2) +
		    pow(PreCA03,2)*pow(errca2[3],2) + pow(PreCA04,2)*pow(errca2[4],2) + pow(PreCA1M1,2)*pow(errca2[8],2) +
		    pow(PreCA10,2)*pow(errca2[9],2) + pow(PreCA11,2)*pow(errca2[10],2) + pow(PreCA12,2)*pow(errca2[11],2) +
		    pow(PreCA13,2)*pow(errca2[12],2) + pow(PreCA00,2)*pow(errca3[0],2) + pow(PreCA01,2)*pow(errca3[1],2) +
		    pow(PreCA02,2)*pow(errca3[2],2) + pow(PreCA03,2)*pow(errca3[3],2) + pow(PrePNF00,2)*pow(errca3[5],2) +
		    pow(PrePNF01,2)*pow(errca3[6],2) + pow(PrePNF02,2)*pow(errca3[7],2) + pow(PreCA1M1,2)*pow(errca3[8],2) +
		    pow(PreCA10,2)*pow(errca3[9],2) + pow(PreCA11,2)*pow(errca3[10],2) + pow(PreCA12,2)*pow(errca3[11],2) +
		    pow(PrePNF1M1,2)*pow(errca3[13],2) + pow(PrePNF10,2)*pow(errca3[14],2) + pow(PrePNF11,2)*pow(errca3[15],2) +
		    pow(PrePNF12,2)*pow(errca3[16],2) + pow(PreCA00,2)*pow(errca4[0],2) + pow(PreCA01,2)*pow(errca4[1],2) +
		    pow(PreCA02,2)*pow(errca4[2],2) + pow(PrePNF00,2)*pow(errca4[5],2) + pow(PrePNF01,2)*pow(errca4[6],2) +
		    pow(PrePNF02,2)*pow(errca4[7],2) + pow(PreCA1M1,2)*pow(errca4[8],2) + pow(PreCA10,2)*pow(errca4[9],2) +
		    pow(PreCA11,2)*pow(errca4[10],2) + pow(PrePNF1M1,2)*pow(errca4[13],2) + pow(PrePNF10,2)*pow(errca4[14],2) +
		    pow(PrePNF11,2)*pow(errca4[15],2) + pow(PreCA00,2)*pow(errca5[0],2) + pow(PreCA01,2)*pow(errca5[1],2) +
		    pow(PrePNF00,2)*pow(errca5[5],2) + pow(PrePNF01,2)*pow(errca5[6],2) + pow(PreCA1M1,2)*pow(errca5[8],2) +
		    pow(PreCA10,2)*pow(errca5[9],2) + pow(PrePNF1M1,2)*pow(errca5[13],2) + pow(PrePNF10,2)*pow(errca5[14],2),0.5);
	relprob=1 - (1 - probca2[0])*(1 - probca2[1])*(1 - probca2[2])*(1 - probca2[3])*(1 - probca2[4])*(1 - probca2[8])*(1 - probca2[9])*
			  (1 - probca2[10])*(1 - probca2[11])*(1 - probca2[12])*(1 - probca3[0])*(1 - probca3[1])*(1 - probca3[2])*(1 - probca3[3])*
			  (1 - probca3[5])*(1 - probca3[6])*(1 - probca3[7])*(1 - probca3[8])*(1 - probca3[9])*(1 - probca3[10])*(1 - probca3[11])*
			  (1 - probca3[13])*(1 - probca3[14])*(1 - probca3[15])*(1 - probca3[16])*(1 - probca4[0])*(1 - probca4[1])*
			  (1 - probca4[2])*(1 - probca4[5])*(1 - probca4[6])*(1 - probca4[7])*(1 - probca4[8])*(1 - probca4[9])*(1 - probca4[10])*
			  (1 - probca4[13])*(1 - probca4[14])*(1 - probca4[15])*(1 - probca5[0])*(1 - probca5[1])*(1 - probca5[5])*
			  (1 - probca5[6])*(1 - probca5[8])*(1 - probca5[9])*(1 - probca5[13])*(1 - probca5[14]);

	cout << "Sufficient data for alpha^(0) eps^(0) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(0) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;
	
	//detach shared memory
	shmdt(ShmRes);
	shmdt(ShmErr);
	shmdt(ShmProb);
	shmdt(ShmNev);
	shmdt(ShmTer);

	//delete/destroy shared memory
	shmctl(ShmResId, IPC_RMID, 0);
	shmctl(ShmErrId, IPC_RMID, 0);
	shmctl(ShmProbId, IPC_RMID, 0);
	shmctl(ShmNevId, IPC_RMID, 0);
	shmctl(ShmTerId, IPC_RMID, 0);
	return 0;
	
	//close output file
	output.close();
}
