/* This function is the main body of the SoftSERVE project for the same-hemisphere contribution to the Cf^2 structure's executable. 
 * Due to the structure of the C language the program essentially creates child processes for each 
 * numerical integration (there's one for each number of dimensions in the integration domain), writes the integrator output to a log file
 * and collects the results of the various integrations. At the end of each integration the numerical results are combined with constant 
 * prefactors to get the coefficients of the bare soft function once the required results are available for each coefficient. The 
 * flow is commented along the flow.
 */ 

#include <Auxiliaries.h>
#include <2Int.h>
#include <2IntAM1.h>
#include <2IntAM2.h>
#include <3Int.h>
#include <3IntAM1.h>
#include <3IntAM2.h>
#include <4Int.h>
#include <4IntAM1.h>
#include <4IntAM2.h>
#include <5Int.h>
#include <5IntAM1.h>
#include <Prefactors.h>
#include <Input_Common.h>
#include <cuba.h>

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

#include <stdlib.h>
#include <sys/types.h>
#include <algorithm>
#include <cstdlib>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/wait.h>

using namespace std;


	/* resn[] contains the results of the n-dim integration for Cf to (alpha, eps)
	 * orders ((0,0),(0,-1),(0,-2),(0,-3),(0,-4),(-1,1),(-1,0),(-1,-1),(-1,-2),
	 * (-1,-3),(-2,2),(-2,1),(-2,0),(-2,-1),(-2,-2))
	 */
	double res2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double res3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double res4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double res5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the error estimates for n-dim Cf according to the scheme outlined above
	double err2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double err3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double err4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double err5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// contains the probabilities that a certain n-dim integration's error estimate is incorrect
	double prob2[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double prob3[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double prob4[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	double prob5[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	/* collected result, error estimate and error probability for a certain eps-order
	 * as soon as the necessary data is available. these variables are only buffers, they are reused
	 * for different eps-orders. The -al/-al2 suffices denotes 1/alpha^x results */
	double result=0;
	double error=0;
	double relprob=0;
	double resultal=0;
	double erroral=0;
	double relprobal=0;
	double resultal2=0;
	double erroral2=0;
	double relprobal2=0;

	//integrator variable containing the number of regions into which the integrator splits the integrand, irrelevant for us
	int nregions=0;

	/* The following variables are only necessary because of a specific problem: The Cuba integrators print their logs
	 * on stdout using puts() and fflush(). I want them to write to a file, but as I don't want to modify the Cuba
	 * sources, I have to work around that. There is a possibility of rerouting stdout to a file using freopen,
	 * but this is irreversible. Once stdout is rerouted, I can't send it back to the console. Solution: let Cuba
	 * run inside a child process that receives a copy of all file descriptors of the parent at its fork. If I
	 * freopen stdout in the child, let it terminate and then continue in the parent, the parent's stdout is still
	 * unaffected. Problem with that:
	 * Child and parent have different memory address spaces, each gets a copy of all variables and works only with
	 * those. So we need to introduce shared memory for the variables that the child modifies and the parent needs to
	 * read, i.e. the result, error estimate and error probability. */

	//process ids for the various child processes
	pid_t child2, child3, child4, child5;

	/*
	 * Shm always denotes something shared
	 */

	//memory ids for shared memory between parent and child (Res = result, Err=error, Prob=probability of incorrect error,
	//Nev=Number of function evaluations, Ter=termination code [i.e. whether or not the program terminated correctly])
	int ShmResId,ShmErrId,ShmProbId,ShmNevId,ShmTerId;
	//pointers to shared memory for results, error estimates and error probabilities (these are arrays of type double)
	double *ShmRes, *ShmErr, *ShmProb;
	//pointer to shared process variable neval (# function evaluations), determined by child, but required by parent
	long long int *ShmNev;
	//pointer to shared process variable terminate, which is determined by child, but required by parent
	int *ShmTer;

	//file stream for output
	ofstream output;
	string outputname;

	 /*     ____             _
	  *    | __ )  ___  __ _(_)_ __    _ __  _ __ ___   __ _ _ __ __ _ _ __ ___
	  *    |  _ \ / _ \/ _` | | '_ \  | '_ \| '__/ _ \ / _` | '__/ _` | '_ ` _ \
	  *    | |_) |  __/ (_| | | | | | | |_) | | | (_) | (_| | | | (_| | | | | | |
	  *    |____/ \___|\__, |_|_| |_| | .__/|_|  \___/ \__, |_|  \__,_|_| |_| |_|
	  *                |___/          |_|              |___/
	  */

int main() {
	//open output file
	outputname="Result "+observable+" Cf A.txt";
	output.open(outputname.c_str());
	
	//Info header
	cout << "Observable name: " << observable << "." << endl;
	cout << "Starting calculation for Cf structure, subregion A." << endl;

	/* Request shared memory for result ("Res", double array of length ncomp [number of computations, i.e. regulator orders]),
	 * error estimate ("Err", double array of length ncomp),
	 * probability of wrong error ("Prob", double array of length ncomp),
	 * neval variable ("Nev", long long int),
	 * termination code ("Ter", int);
	 *
	 * shmget returns the ID of the memory block (int type) if successful, -1 if it fails */
	ShmResId=shmget(IPC_PRIVATE, ncomp*sizeof(double), 0777|IPC_CREAT);
	ShmErrId=shmget(IPC_PRIVATE, ncomp*sizeof(double), 0777|IPC_CREAT);
	ShmProbId=shmget(IPC_PRIVATE, ncomp*sizeof(double), 0777|IPC_CREAT);
	ShmNevId=shmget(IPC_PRIVATE, sizeof(long long int), 0777|IPC_CREAT);
	ShmTerId=shmget(IPC_PRIVATE, sizeof(int), 0777|IPC_CREAT);

	//check if shared memory request was successful
	if(ShmResId<0 || ShmErrId <0 || ShmProbId<0 || ShmNevId<0 || ShmTerId<0) {
		cout << "Shared memory request failed." << endl;
		exit(EXIT_FAILURE);
	}

	/* Attach memory to parent's address space
	 *
	 * This construction is necessary, as shmat returns a pointer of type void*, which should be typecast
	 * to whatever we need (double* for ShmRes, ShmErr, ShmProb; long long int* for ShmNev; and int* for ShmTer).
	 * However, if shmat fails, it returns a char* -1, so to check I assign this return value to a void* pointer
	 * called Attx, check if it evaluates to -1 if typecast as char*, and if it doesn't (i.e. if the attachment
	 * was successful), I typecast Attx to double*, long long int*, or int*, and assign it to ShmRes, etc.	*/
	void *Att1,*Att2,*Att3,*Att4,*Att5;
	Att1 = shmat(ShmResId, NULL, 0);
	Att2 = shmat(ShmErrId, NULL, 0);
	Att3 = shmat(ShmProbId, NULL, 0);
	Att4 = shmat(ShmNevId, NULL, 0);
	Att5 = shmat(ShmTerId, NULL, 0);
	if((char*)Att1==(char*)-1 || (char*)Att2==(char*)-1 || (char*)Att3==(char*)-1 || (char*)Att4==(char*)-1 || (char*)Att5==(char*)-1) {
		cout << "Shared memory attachment failed" << endl;
		exit(EXIT_FAILURE);
	} else {
		ShmRes = (double*) Att1;
		ShmErr = (double*) Att2;
		ShmProb = (double*) Att3;
		ShmNev = (long long int*) Att4;
		ShmTer = (int*) Att5;
	}

	//finally, start first integration
	cout << "2-dimensional integration in progress..." << endl;
	//create child process and check if this was successful
	child2=fork();
	if(child2 <0) { ///
		cout << "Fork for 2-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child2==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncomp; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources (D2I = 'D'imension'2' 'I'ntegration)
		integrand_t D2I= &D2Int;

		//reroute output of llDivonne to a text file
		if(freopen("2-dimensional CFA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim2, ncomp, D2I,0, 1, epsrel2, epsabs2, flags, seed, 0, maxeval2, key12, key22, key32, maxpass2, border, maxchisq, mindev, 0, ndim2, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before the parent does anything */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncomp),res2);
	copy(ShmErr,(ShmErr+ncomp),err2);
	copy(ShmProb,(ShmProb+ncomp),prob2);

	//first integration finished
	if(*ShmTer<0) {
		cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
	} else if(*ShmTer>0) {
		cout << " 2-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
	} else if(*ShmTer==0){
		cout << "2-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
	} else {
		cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
	}

	//compute first result (2-dim integration alone is sufficient for div^(-4) coefficient)
	result=PreE0A0*res2[4] + PreEM1A1*res2[9] + PreEM2A2*res2[14];
	error=pow(pow(PreE0A0,2)*pow(err2[4],2) + pow(PreEM1A1,2)*pow(err2[9],2) +
		    pow(PreEM2A2,2)*pow(err2[14],2),0.5);
	relprob=1 - (1 - prob2[4])*(1 - prob2[9])*(1 - prob2[14]);

	resultal=PreE0A0*res2[9] + PreEM1A1*res2[14];
	erroral=pow(pow(PreE0A0,2)*pow(err2[9],2) + pow(PreEM1A1,2)*pow(err2[14],2),0.5);
	relprobal=1 - (1 - prob2[9])*(1 - prob2[14]);

	resultal2=PreE0A0*res2[14];
	erroral2=pow(pow(PreE0A0,2)*pow(err2[14],2),0.5);
	relprobal2=prob2[14];

	//announce the result to the world at large
	cout << "Sufficient data for alpha^(0) eps^(-4) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-4) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-3) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-3) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;
	cout << "Sufficient data for alpha^(-2) eps^(-2) coefficient available, result: " << endl << resultal2 << " +/- " << erroral2 << ", with probability of incorrect error estimate of " << relprobal2 << "." <<  endl;
	output << "alpha^(-2) eps^(-2) coefficient: " << resultal2 << " +/- " << erroral2 << ", probability of incorrect error estimate: " << relprobal2 << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;
	resultal2=0;
	erroral2=0;
	relprobal2=0;

	//start second integration
	cout << "3-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child3=fork();
	if(child3 <0) { ///
		cout << "Fork for 3-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child3==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncomp; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D3I= &D3Int;

		//reroute output of llDivonne to a text file
		if(freopen("3-dimensional CFA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim3, ncomp, D3I,0, 1, epsrel3, epsabs3, flags, seed, 0, maxeval3, key13, key23, key33, maxpass3, border, maxchisq, mindev, 0, ndim3, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncomp),res3);
	copy(ShmErr,(ShmErr+ncomp),err3);
	copy(ShmProb,(ShmProb+ncomp),prob3);

	//second integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 3-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "3-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}

	//compute second result (div^(-3) needs only 2-D and 3-D)
	result=PreE0A0*res2[3] + PreE1A0*res2[4] + PreEM1A1*res2[8] + PreE0A1*res2[9] +
			   PreEM2A2*res2[13] + PreEM1A2*res2[14] + PreE0A0*res3[3] + PreEM1A1*res3[8] +
			   PreEM2A2*res3[13];
	error=pow(pow(PreE0A0,2)*pow(err2[3],2) + pow(PreE1A0,2)*pow(err2[4],2) +
		    pow(PreEM1A1,2)*pow(err2[8],2) + pow(PreE0A1,2)*pow(err2[9],2) +
		    pow(PreEM2A2,2)*pow(err2[13],2) + pow(PreEM1A2,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[3],2) + pow(PreEM1A1,2)*pow(err3[8],2) +
		    pow(PreEM2A2,2)*pow(err3[13],2),0.5);
	relprob=1 - (1 - prob2[3])*(1 - prob2[4])*(1 - prob2[8])*(1 - prob2[9])*(1 - prob2[13])*
		    (1 - prob2[14])*(1 - prob3[3])*(1 - prob3[8])*(1 - prob3[13]);

	resultal=PreE0A0*res2[8] + PreE1A0*res2[9] + PreEM1A1*res2[13] + PreE0A1*res2[14] +
			   PreE0A0*res3[8] + PreEM1A1*res3[13];
	erroral=pow(pow(PreE0A0,2)*pow(err2[8],2) + pow(PreE1A0,2)*pow(err2[9],2) +
		    pow(PreEM1A1,2)*pow(err2[13],2) + pow(PreE0A1,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[8],2) + pow(PreEM1A1,2)*pow(err3[13],2),0.5);
	relprobal=1 - (1 - prob2[8])*(1 - prob2[9])*(1 - prob2[13])*(1 - prob2[14])*(1 - prob3[8])*
		    (1 - prob3[13]);

	resultal2=PreE0A0*res2[13] + PreE1A0*res2[14] + PreE0A0*res3[13];
	erroral2=pow(pow(PreE0A0,2)*pow(err2[13],2) + pow(PreE1A0,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[13],2),0.5);
	relprobal2=1 - (1 - prob2[13])*(1 - prob2[14])*(1 - prob3[13]);

	cout << "Sufficient data for alpha^(0) eps^(-3) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-3) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-2) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-2) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;
	cout << "Sufficient data for alpha^(-2) eps^(-1) coefficient available, result: " << endl << resultal2 << " +/- " << erroral2 << ", with probability of incorrect error estimate of " << relprobal2 << "." <<  endl;
	output << "alpha^(-2) eps^(-1) coefficient: " << resultal2 << " +/- " << erroral2 << ", probability of incorrect error estimate: " << relprobal2 << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;
	resultal2=0;
	erroral2=0;
	relprobal2=0;

	//start third integration
	cout << "4-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child4=fork();
	if(child4 <0) { ///
		cout << "Fork for 4-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child4==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncomp; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D4I= &D4Int;

		//reroute output of llDivonne to a text file
		if(freopen("4-dimensional CFA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim4, ncomp, D4I,0, 1, epsrel4, epsabs4, flags, seed, 0, maxeval4, key14, key24, key34, maxpass4, border, maxchisq, mindev, 0, ndim4, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncomp),res4);
	copy(ShmErr,(ShmErr+ncomp),err4);
	copy(ShmProb,(ShmProb+ncomp),prob4);

	//third integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 4-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "4-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}
	//compute third result (div^(-2) needs 2-D, 3-D and 4-D)
	result=PreE0A0*res2[2] + PreE1A0*res2[3] + PreE2A0*res2[4] + PreEM1A1*res2[7] + 
			   PreE0A1*res2[8] + PreE1A1*res2[9] + PreEM2A2*res2[12] + PreEM1A2*res2[13] +
			   PreE0A2*res2[14] + PreE0A0*res3[2] + PreE1A0*res3[3] + PreEM1A1*res3[7] +
			   PreE0A1*res3[8] + PreEM2A2*res3[12] + PreEM1A2*res3[13] + PreE0A0*res4[2] +
			   PreEM1A1*res4[7];
	error=pow(pow(PreE0A0,2)*pow(err2[2],2) + pow(PreE1A0,2)*pow(err2[3],2) +
		    pow(PreE2A0,2)*pow(err2[4],2) + pow(PreEM1A1,2)*pow(err2[7],2) +
		    pow(PreE0A1,2)*pow(err2[8],2) + pow(PreE1A1,2)*pow(err2[9],2) +
		    pow(PreEM2A2,2)*pow(err2[12],2) + pow(PreEM1A2,2)*pow(err2[13],2) +
		    pow(PreE0A2,2)*pow(err2[14],2) + pow(PreE0A0,2)*pow(err3[2],2) +
		    pow(PreE1A0,2)*pow(err3[3],2) + pow(PreEM1A1,2)*pow(err3[7],2) +
		    pow(PreE0A1,2)*pow(err3[8],2) + pow(PreEM2A2,2)*pow(err3[12],2) +
		    pow(PreEM1A2,2)*pow(err3[13],2) + pow(PreE0A0,2)*pow(err4[2],2) +
		    pow(PreEM1A1,2)*pow(err4[7],2),0.5);
	relprob=1 - (1 - prob2[2])*(1 - prob2[3])*(1 - prob2[4])*(1 - prob2[7])*(1 - prob2[8])*
		    (1 - prob2[9])*(1 - prob2[12])*(1 - prob2[13])*(1 - prob2[14])*(1 - prob3[2])*
		    (1 - prob3[3])*(1 - prob3[7])*(1 - prob3[8])*(1 - prob3[12])*(1 - prob3[13])*
		    (1 - prob4[2])*(1 - prob4[7]);

	resultal=PreE0A0*res2[7] + PreE1A0*res2[8] + PreE2A0*res2[9] + PreEM1A1*res2[12] + 
			   PreE0A1*res2[13] + PreE1A1*res2[14] + PreE0A0*res3[7] + PreE1A0*res3[8] +
			   PreEM1A1*res3[12] + PreE0A1*res3[13] + PreE0A0*res4[7];
	erroral=pow(pow(PreE0A0,2)*pow(err2[7],2) + pow(PreE1A0,2)*pow(err2[8],2) +
		    pow(PreE2A0,2)*pow(err2[9],2) + pow(PreEM1A1,2)*pow(err2[12],2) +
		    pow(PreE0A1,2)*pow(err2[13],2) + pow(PreE1A1,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[7],2) + pow(PreE1A0,2)*pow(err3[8],2) +
		    pow(PreEM1A1,2)*pow(err3[12],2) + pow(PreE0A1,2)*pow(err3[13],2) +
		    pow(PreE0A0,2)*pow(err4[7],2),0.5);
	relprobal=1 - (1 - prob2[7])*(1 - prob2[8])*(1 - prob2[9])*(1 - prob2[12])*(1 - prob2[13])*
		    (1 - prob2[14])*(1 - prob3[7])*(1 - prob3[8])*(1 - prob3[12])*(1 - prob3[13])*
		    (1 - prob4[7]);

	resultal2=PreE0A0*res2[12] + PreE1A0*res2[13] + PreE2A0*res2[14] + PreE0A0*res3[12] + 
			   PreE1A0*res3[13];
	erroral2=pow(pow(PreE0A0,2)*pow(err2[12],2) + pow(PreE1A0,2)*pow(err2[13],2) +
		    pow(PreE2A0,2)*pow(err2[14],2) + pow(PreE0A0,2)*pow(err3[12],2) +
		    pow(PreE1A0,2)*pow(err3[13],2),0.5);
	relprobal2=1 - (1 - prob2[12])*(1 - prob2[13])*(1 - prob2[14])*(1 - prob3[12])*(1 - prob3[13]);

	cout << "Sufficient data for alpha^(0) eps^(-2) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-2) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(-1) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(-1) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;
	cout << "Sufficient data for alpha^(-2) eps^(0) coefficient available, result: " << endl << resultal2 << " +/- " << erroral2 << ", with probability of incorrect error estimate of " << relprobal2 << "." <<  endl;
	output << "alpha^(-2) eps^(0) coefficient: " << resultal2 << " +/- " << erroral2 << ", probability of incorrect error estimate: " << relprobal2 << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;
	resultal2=0;
	erroral2=0;
	relprobal2=0;

	//start fourth integration
	cout << "5-dimensional integration in progress..." << endl;

	//create child process and check if this was successful
	child5=fork();
	if(child5 <0) { ///
		cout << "Fork for 5-dimensional integration failed" << endl;
		exit(EXIT_FAILURE);
	} //everything in here is the child process
	else if(child5==0) {

		//reset shared memory to zeros...
		for(int i=0; i<ncomp; i++) {
				*(ShmRes+i)=0;
				*(ShmErr+i)=0;
				*(ShmProb+i)=0;
			}

		//choose the integrand from the auxiliary sources
		integrand_t D5I= &D5Int;

		//reroute output of llDivonne to a text file
		if(freopen("5-dimensional CFA.log","w",stdout)==NULL) {
			cout << "Redirecting output stream failed" << endl;
			exit(EXIT_FAILURE);
		}

		//perform the integration
		llDivonne(ndim5, ncomp, D5I,0, 1, epsrel5, epsabs5, flags, seed, 0, maxeval5, key15, key25, key35, maxpass5, border, maxchisq, mindev, 0, ndim5, 0, 0, 0, "", 0, &nregions, ShmNev, ShmTer, ShmRes, ShmErr, ShmProb);

		//terminate the child
		_exit(EXIT_SUCCESS);
	}

	/* Parent process
	 *
	 * Strictly, this is shared code between child and parent, but the child is terminated well before this point */

	//wait until child has terminated
	wait(NULL);

	//transfer results from shared memory to parent
	copy(ShmRes,(ShmRes+ncomp),res5);
	copy(ShmErr,(ShmErr+ncomp),err5);
	copy(ShmProb,(ShmProb+ncomp),prob5);

	//fourth integration finished
	if(*ShmTer<0) {
			cout << "Dimension of integration out of range, please contact developer, this should never happen." << endl;
		} else if(*ShmTer>0) {
			cout << " 5-d integration stopped, accuracy not reached. Continuing with intermediate results." << endl;
		} else if(*ShmTer==0){
			cout << "5-d integration successful after " << *ShmNev << " function evaluations." <<  endl;
		} else {
			cout << "How did you manage to get here?? Something REALLY went wrong..." << endl;
		}

	//compute last results (div^(-1) and div^0 need 2-D to 5-D; 6-D is not required for div^0, counterintuitively, here we compute div^(-1))
	result=PreE0A0*res2[1] + PreE1A0*res2[2] + PreE2A0*res2[3] + PreE3A0*res2[4] + 
			   PreEM1A1*res2[6] + PreE0A1*res2[7] + PreE1A1*res2[8] + PreE2A1*res2[9] +
			   PreEM2A2*res2[11] + PreEM1A2*res2[12] + PreE0A2*res2[13] + PreE1A2*res2[14] +
			   PreE0A0*res3[1] + PreE1A0*res3[2] + PreE2A0*res3[3] + PreEM1A1*res3[6] +
			   PreE0A1*res3[7] + PreE1A1*res3[8] + PreEM2A2*res3[11] + PreEM1A2*res3[12] +
			   PreE0A2*res3[13] + PreE0A0*res4[1] + PreE1A0*res4[2] + PreEM1A1*res4[6] +
			   PreE0A1*res4[7] + PreEM2A2*res4[11] + PreE0A0*res5[1] + PreEM1A1*res5[6];
	error=pow(pow(PreE0A0,2)*pow(err2[1],2) + pow(PreE1A0,2)*pow(err2[2],2) +
		    pow(PreE2A0,2)*pow(err2[3],2) + pow(PreE3A0,2)*pow(err2[4],2) +
		    pow(PreEM1A1,2)*pow(err2[6],2) + pow(PreE0A1,2)*pow(err2[7],2) +
		    pow(PreE1A1,2)*pow(err2[8],2) + pow(PreE2A1,2)*pow(err2[9],2) +
		    pow(PreEM2A2,2)*pow(err2[11],2) + pow(PreEM1A2,2)*pow(err2[12],2) +
		    pow(PreE0A2,2)*pow(err2[13],2) + pow(PreE1A2,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[1],2) + pow(PreE1A0,2)*pow(err3[2],2) +
		    pow(PreE2A0,2)*pow(err3[3],2) + pow(PreEM1A1,2)*pow(err3[6],2) +
		    pow(PreE0A1,2)*pow(err3[7],2) + pow(PreE1A1,2)*pow(err3[8],2) +
		    pow(PreEM2A2,2)*pow(err3[11],2) + pow(PreEM1A2,2)*pow(err3[12],2) +
		    pow(PreE0A2,2)*pow(err3[13],2) + pow(PreE0A0,2)*pow(err4[1],2) +
		    pow(PreE1A0,2)*pow(err4[2],2) + pow(PreEM1A1,2)*pow(err4[6],2) +
		    pow(PreE0A1,2)*pow(err4[7],2) + pow(PreEM2A2,2)*pow(err4[11],2) +
		    pow(PreE0A0,2)*pow(err5[1],2) + pow(PreEM1A1,2)*pow(err5[6],2),0.5);
	relprob=1 - (1 - prob2[1])*(1 - prob2[2])*(1 - prob2[3])*(1 - prob2[4])*(1 - prob2[6])*
		    (1 - prob2[7])*(1 - prob2[8])*(1 - prob2[9])*(1 - prob2[11])*(1 - prob2[12])*
		    (1 - prob2[13])*(1 - prob2[14])*(1 - prob3[1])*(1 - prob3[2])*(1 - prob3[3])*
		    (1 - prob3[6])*(1 - prob3[7])*(1 - prob3[8])*(1 - prob3[11])*(1 - prob3[12])*
		    (1 - prob3[13])*(1 - prob4[1])*(1 - prob4[2])*(1 - prob4[6])*(1 - prob4[7])*
		    (1 - prob4[11])*(1 - prob5[1])*(1 - prob5[6]);

	resultal=PreE0A0*res2[6] + PreE1A0*res2[7] + PreE2A0*res2[8] + PreE3A0*res2[9] + 
			   PreEM1A1*res2[11] + PreE0A1*res2[12] + PreE1A1*res2[13] + PreE2A1*res2[14] +
			   PreE0A0*res3[6] + PreE1A0*res3[7] + PreE2A0*res3[8] + PreEM1A1*res3[11] +
			   PreE0A1*res3[12] + PreE1A1*res3[13] + PreE0A0*res4[6] + PreE1A0*res4[7] +
			   PreEM1A1*res4[11] + PreE0A0*res5[6];
	erroral=pow(pow(PreE0A0,2)*pow(err2[6],2) + pow(PreE1A0,2)*pow(err2[7],2) +
		    pow(PreE2A0,2)*pow(err2[8],2) + pow(PreE3A0,2)*pow(err2[9],2) +
		    pow(PreEM1A1,2)*pow(err2[11],2) + pow(PreE0A1,2)*pow(err2[12],2) +
		    pow(PreE1A1,2)*pow(err2[13],2) + pow(PreE2A1,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err3[6],2) + pow(PreE1A0,2)*pow(err3[7],2) +
		    pow(PreE2A0,2)*pow(err3[8],2) + pow(PreEM1A1,2)*pow(err3[11],2) +
		    pow(PreE0A1,2)*pow(err3[12],2) + pow(PreE1A1,2)*pow(err3[13],2) +
		    pow(PreE0A0,2)*pow(err4[6],2) + pow(PreE1A0,2)*pow(err4[7],2) +
		    pow(PreEM1A1,2)*pow(err4[11],2) + pow(PreE0A0,2)*pow(err5[6],2),0.5);
	relprobal=1 - (1 - prob2[6])*(1 - prob2[7])*(1 - prob2[8])*(1 - prob2[9])*(1 - prob2[11])*
		    (1 - prob2[12])*(1 - prob2[13])*(1 - prob2[14])*(1 - prob3[6])*(1 - prob3[7])*
		    (1 - prob3[8])*(1 - prob3[11])*(1 - prob3[12])*(1 - prob3[13])*(1 - prob4[6])*
		    (1 - prob4[7])*(1 - prob4[11])*(1 - prob5[6]);

	cout << "Sufficient data for alpha^(0) eps^(-1) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(-1) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;
	cout << "Sufficient data for alpha^(-1) eps^(0) coefficient available, result: " << endl << resultal << " +/- " << erroral << ", with probability of incorrect error estimate of " << relprobal << "." <<  endl;
	output << "alpha^(-1) eps^(0) coefficient: " << resultal << " +/- " << erroral << ", probability of incorrect error estimate: " << relprobal << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;
	resultal=0;
	erroral=0;
	relprobal=0;	

	// final result for (0,0) order
	result=PreE1A0*res2[1] + PreE2A0*res2[2] + PreE3A0*res2[3] + PreE4A0*res2[4] + 
			   PreEM1A1*res2[5] + PreE0A1*res2[6] + PreE1A1*res2[7] + PreE2A1*res2[8] +
			   PreE3A1*res2[9] + PreEM2A2*res2[10] + PreEM1A2*res2[11] + PreE0A2*res2[12] +
			   PreE1A2*res2[13] + PreE2A2*res2[14] + PreE0A0*res2[0] + PreE1A0*res3[1] +
			   PreE2A0*res3[2] + PreE3A0*res3[3] + PreEM1A1*res3[5] + PreE0A1*res3[6] +
			   PreE1A1*res3[7] + PreE2A1*res3[8] + PreEM2A2*res3[10] + PreEM1A2*res3[11] +
			   PreE0A2*res3[12] + PreE1A2*res3[13] + PreE0A0*res3[0] + PreE1A0*res4[1] +
			   PreE2A0*res4[2] + PreEM1A1*res4[5] + PreE0A1*res4[6] + PreE1A1*res4[7] +
			   PreEM2A2*res4[10] + PreEM1A2*res4[11] + PreE0A0*res4[0] + PreE1A0*res5[1] +
			   PreE0A1*res5[6] + PreE0A0*res5[0];
	error=pow(pow(PreE1A0,2)*pow(err2[1],2) + pow(PreE2A0,2)*pow(err2[2],2) +
		    pow(PreE3A0,2)*pow(err2[3],2) + pow(PreE4A0,2)*pow(err2[4],2) +
		    pow(PreEM1A1,2)*pow(err2[5],2) + pow(PreE0A1,2)*pow(err2[6],2) +
		    pow(PreE1A1,2)*pow(err2[7],2) + pow(PreE2A1,2)*pow(err2[8],2) +
		    pow(PreE3A1,2)*pow(err2[9],2) + pow(PreEM2A2,2)*pow(err2[10],2) +
		    pow(PreEM1A2,2)*pow(err2[11],2) + pow(PreE0A2,2)*pow(err2[12],2) +
		    pow(PreE1A2,2)*pow(err2[13],2) + pow(PreE2A2,2)*pow(err2[14],2) +
		    pow(PreE0A0,2)*pow(err2[0],2) + pow(PreE1A0,2)*pow(err3[1],2) +
		    pow(PreE2A0,2)*pow(err3[2],2) + pow(PreE3A0,2)*pow(err3[3],2) +
		    pow(PreEM1A1,2)*pow(err3[5],2) + pow(PreE0A1,2)*pow(err3[6],2) +
		    pow(PreE1A1,2)*pow(err3[7],2) + pow(PreE2A1,2)*pow(err3[8],2) +
		    pow(PreEM2A2,2)*pow(err3[10],2) + pow(PreEM1A2,2)*pow(err3[11],2) +
		    pow(PreE0A2,2)*pow(err3[12],2) + pow(PreE1A2,2)*pow(err3[13],2) +
		    pow(PreE0A0,2)*pow(err3[0],2) + pow(PreE1A0,2)*pow(err4[1],2) +
		    pow(PreE2A0,2)*pow(err4[2],2) + pow(PreEM1A1,2)*pow(err4[5],2) +
		    pow(PreE0A1,2)*pow(err4[6],2) + pow(PreE1A1,2)*pow(err4[7],2) +
		    pow(PreEM2A2,2)*pow(err4[10],2) + pow(PreEM1A2,2)*pow(err4[11],2) +
		    pow(PreE0A0,2)*pow(err4[0],2) + pow(PreE1A0,2)*pow(err5[1],2) +
		    pow(PreE0A1,2)*pow(err5[6],2) + pow(PreE0A0,2)*pow(err5[0],2),0.5);
	relprob=1 - (1 - prob2[1])*(1 - prob2[2])*(1 - prob2[3])*(1 - prob2[4])*(1 - prob2[5])*
		    (1 - prob2[6])*(1 - prob2[7])*(1 - prob2[8])*(1 - prob2[9])*(1 - prob2[10])*
		    (1 - prob2[11])*(1 - prob2[12])*(1 - prob2[13])*(1 - prob2[14])*(1 - prob2[0])*
		    (1 - prob3[1])*(1 - prob3[2])*(1 - prob3[3])*(1 - prob3[5])*(1 - prob3[6])*
		    (1 - prob3[7])*(1 - prob3[8])*(1 - prob3[10])*(1 - prob3[11])*(1 - prob3[12])*
		    (1 - prob3[13])*(1 - prob3[0])*(1 - prob4[1])*(1 - prob4[2])*(1 - prob4[5])*
		    (1 - prob4[6])*(1 - prob4[7])*(1 - prob4[10])*(1 - prob4[11])*(1 - prob4[0])*
		    (1 - prob5[1])*(1 - prob5[6])*(1 - prob5[0]);

	cout << "Sufficient data for alpha^(0) eps^(0) coefficient available, result: " << endl << result << " +/- " << error << ", with probability of incorrect error estimate of " << relprob << "." <<  endl;
	output << "alpha^(0) eps^(0) coefficient: " << result << " +/- " << error << ", probability of incorrect error estimate: " << relprob << endl;

	//clear results buffer
	result=0;
	error=0;
	relprob=0;

	//detach shared memory
	shmdt(ShmRes);
	shmdt(ShmErr);
	shmdt(ShmProb);
	shmdt(ShmNev);
	shmdt(ShmTer);

	//delete/destroy shared memory
	shmctl(ShmResId, IPC_RMID, 0);
	shmctl(ShmErrId, IPC_RMID, 0);
	shmctl(ShmProbId, IPC_RMID, 0);
	shmctl(ShmNevId, IPC_RMID, 0);
	shmctl(ShmTerId, IPC_RMID, 0);
	return 0;
	
	//close output file
	output.close();
}
