/*
 * This file contains the difference structures arising from plus distributions tying together the Measurement functions FXi, plus distribution denominators 
 * and contributing remainder/numerator functions XCfEiAj for the 4D integration at all eps-orders and the alpha^0 order of the Cf^2 structure, subregion b. 
 * Example: I4EM1A0kl ('I'ntegration over '4' dimensions, at 'E'ps to the 'M'inus '1'st and 'A'lpha to the '0'th order, third and fourth integrations over 
 * x'k' and x'l') contains various logs of FXi functions evaluated at c=0,s5=0, i.e. with the two deltas from the c and s5 +-distributions, multiplied 
 * with the correct X functions' eps- and alpha-powers to arrive at eps^-1 alpha^0. 
 */

#include <4Int.h>
#include <Auxiliaries.h>
#include <cmath>
#include <XFactor.h>
#include <Input_Common.h>

using namespace std;

double I4EM2A0kl(double xk, double xl, double s, double s6) {
	return 0; //no alpha-poles, but leading order is xi-indep
}
double I4EM1A0kl(double xk, double xl, double s, double s6) {
	return (pow(xk,-1)*pow(xl,-1)*(2*(log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,yl(xl),t(s),0,tl(s6),0)) - log(FB1(yk(xk),0,t(s),0,tl(s6),0)) + log(FB1(yk(xk),yl(xl),t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),0)) -
	          log(FB2(0,yl(xl),t(s),0,tl(s6),0)) - log(FB2(yk(xk),0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),yl(xl),t(s),0,tl(s6),0)))*XCfE0A0(s,0,s6) + XCfE1A0(0,0,s,0,s6) - XCfE1A0(0,xl,s,0,s6) - XCfE1A0(xk,0,s,0,s6) +
	       XCfE1A0(xk,xl,s,0,s6)))/4.;
}
double I4E0A0kl(double xk, double xl, double s, double s6) {
	return (pow(xk,-1)*pow(xl,-1)*(4*(pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(0,yl(xl),t(s),0,tl(s6),0)),2) - pow(log(FB1(yk(xk),0,t(s),0,tl(s6),0)),2) + pow(log(FB1(yk(xk),yl(xl),t(s),0,tl(s6),0)),2) +
	          pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(0,yl(xl),t(s),0,tl(s6),0)),2) - pow(log(FB2(yk(xk),0,t(s),0,tl(s6),0)),2) + pow(log(FB2(yk(xk),yl(xl),t(s),0,tl(s6),0)),2))*XCfE0A0(s,0,s6) +
	       2*(log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),0)))*XCfE1A0(0,0,s,0,s6) - 2*(log(FB1(0,yl(xl),t(s),0,tl(s6),0)) + log(FB2(0,yl(xl),t(s),0,tl(s6),0)))*XCfE1A0(0,xl,s,0,s6) -
	       2*(log(FB1(yk(xk),0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),0,t(s),0,tl(s6),0)))*XCfE1A0(xk,0,s,0,s6) +
	       2*(log(FB1(yk(xk),yl(xl),t(s),0,tl(s6),0)) + log(FB2(yk(xk),yl(xl),t(s),0,tl(s6),0)))*XCfE1A0(xk,xl,s,0,s6) + XCfE2A0(0,0,s,0,s6) - XCfE2A0(0,xl,s,0,s6) - XCfE2A0(xk,0,s,0,s6) + XCfE2A0(xk,xl,s,0,s6)))/4.;
}

double I4EM2A0kc(double c, double xk, double s, double s6) {
	return 0; //one alpha pole, which can be cancelled, but not converted
}
double I4EM1A0kc(double c, double xk, double s, double s6) {
	return -2*(log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) - log(FB1(yk(xk),0,t(s),0,tl(s6),0)) + log(FB1(yk(xk),0,t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) -
		     log(FB2(0,0,t(s),0,tl(s6),b(c))) - log(FB2(yk(xk),0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),0,t(s),0,tl(s6),b(c))))*pow(c,-1)*pow(mexp,-1)*pow(xk,-1)*XCfE0A0(s,0,s6);
}
double I4E0A0kc(double c, double xk, double s, double s6) {
	return 2*pow(c,-1)*pow(mexp,-1)*pow(xk,-1)*((mexp*log(xk)*(-log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB1(0,0,t(s),0,tl(s6),b(c))) + log(FB1(yk(xk),0,t(s),0,tl(s6),0)) - log(FB1(yk(xk),0,t(s),0,tl(s6),b(c))) -
	           log(FB2(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),b(c))) + log(FB2(yk(xk),0,t(s),0,tl(s6),0)) - log(FB2(yk(xk),0,t(s),0,tl(s6),b(c)))) +
	        8*log(c)*(log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) - log(FB1(yk(xk),0,t(s),0,tl(s6),0)) + log(FB1(yk(xk),0,t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) -
	           log(FB2(0,0,t(s),0,tl(s6),b(c))) - log(FB2(yk(xk),0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),0,t(s),0,tl(s6),b(c)))) -
	        4*(pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(0,0,t(s),0,tl(s6),b(c))),2) - pow(log(FB1(yk(xk),0,t(s),0,tl(s6),0)),2) + pow(log(FB1(yk(xk),0,t(s),0,tl(s6),b(c))),2) +
	           pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(0,0,t(s),0,tl(s6),b(c))),2) - pow(log(FB2(yk(xk),0,t(s),0,tl(s6),0)),2) + pow(log(FB2(yk(xk),0,t(s),0,tl(s6),b(c))),2)))*XCfE0A0(s,0,s6) -
	     (log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) - log(FB2(0,0,t(s),0,tl(s6),b(c))))*(2*XCfE0A1(0,0,s,0,s6) + XCfE1A0(0,0,s,0,s6)) +
	     (log(FB1(yk(xk),0,t(s),0,tl(s6),0)) - log(FB1(yk(xk),0,t(s),0,tl(s6),b(c))) + log(FB2(yk(xk),0,t(s),0,tl(s6),0)) - log(FB2(yk(xk),0,t(s),0,tl(s6),b(c))))*(2*XCfE0A1(xk,0,s,0,s6) + XCfE1A0(xk,0,s,0,s6)));
}

double I4EM2A0ks(double xk, double s, double s5, double s6) {
	return 0; // one alpha pole, but leading order is xk-indep
}
double I4EM1A0ks(double xk, double s, double s5, double s6) {
	return (pow(mexp,-1)*pow(s5,-1)*pow(xk,-1)*(-2*XCfE0A1(0,0,s,0,s6) + 2*XCfE0A1(0,0,s,s5,s6) + 2*XCfE0A1(xk,0,s,0,s6) - 2*XCfE0A1(xk,0,s,s5,s6) + XCfE1A0(0,0,s,0,s6) - XCfE1A0(0,0,s,s5,s6) - XCfE1A0(xk,0,s,0,s6) +
		       XCfE1A0(xk,0,s,s5,s6)))/2.;
}
double I4E0A0ks(double xk, double s, double s5, double s6) {
	return (pow(mexp,-1)*pow(s5,-1)*pow(xk,-1)*(2*(mexp*log(xk)*(-log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB1(yk(xk),0,t(s),0,tl(s6),0)) - log(FB2(0,0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),0,t(s),0,tl(s6),0))) -
	          2*(pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(yk(xk),0,t(s),0,tl(s6),0)),2) + pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(yk(xk),0,t(s),0,tl(s6),0)),2)))*XCfE0A0(s,0,s6) +
	       2*(mexp*log(xk)*(log(FB1(0,0,t(s),t5(s5),tl(s6),0)) - log(FB1(yk(xk),0,t(s),t5(s5),tl(s6),0)) + log(FB2(0,0,t(s),t5(s5),tl(s6),0)) - log(FB2(yk(xk),0,t(s),t5(s5),tl(s6),0))) +
	          2*(pow(log(FB1(0,0,t(s),t5(s5),tl(s6),0)),2) - pow(log(FB1(yk(xk),0,t(s),t5(s5),tl(s6),0)),2) + pow(log(FB2(0,0,t(s),t5(s5),tl(s6),0)),2) - pow(log(FB2(yk(xk),0,t(s),t5(s5),tl(s6),0)),2)))*
	        XCfE0A0(s,s5,s6) - 4*(log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),0)))*XCfE0A1(0,0,s,0,s6) + 4*(log(FB1(0,0,t(s),t5(s5),tl(s6),0)) + log(FB2(0,0,t(s),t5(s5),tl(s6),0)))*XCfE0A1(0,0,s,s5,s6) +
	       4*(log(FB1(yk(xk),0,t(s),0,tl(s6),0)) + log(FB2(yk(xk),0,t(s),0,tl(s6),0)))*XCfE0A1(xk,0,s,0,s6) -
	       4*(log(FB1(yk(xk),0,t(s),t5(s5),tl(s6),0)) + log(FB2(yk(xk),0,t(s),t5(s5),tl(s6),0)))*XCfE0A1(xk,0,s,s5,s6) +
	       mexp*log(xk)*(-XCfE1A0(0,0,s,0,s6) + XCfE1A0(0,0,s,s5,s6) + XCfE1A0(xk,0,s,0,s6) - XCfE1A0(xk,0,s,s5,s6)) +
	       2*log(s5)*(2*XCfE0A1(0,0,s,0,s6) - 2*XCfE0A1(0,0,s,s5,s6) - 2*XCfE0A1(xk,0,s,0,s6) + 2*XCfE0A1(xk,0,s,s5,s6) - XCfE1A0(0,0,s,0,s6) + XCfE1A0(0,0,s,s5,s6) + XCfE1A0(xk,0,s,0,s6) - XCfE1A0(xk,0,s,s5,s6)) -
	       2*XCfE1A1(0,0,s,0,s6) + 2*XCfE1A1(0,0,s,s5,s6) + 2*XCfE1A1(xk,0,s,0,s6) - 2*XCfE1A1(xk,0,s,s5,s6) + XCfE2A0(0,0,s,0,s6) - XCfE2A0(0,0,s,s5,s6) - XCfE2A0(xk,0,s,0,s6) + XCfE2A0(xk,0,s,s5,s6)))/2.;
}

double I4EM2A0lc(double c, double xl, double s, double s6) {
	return 0; //leading order is xl-indep
}
double I4EM1A0lc(double c, double xl, double s, double s6) {
	return -2*(log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) - log(FB1(0,yl(xl),t(s),0,tl(s6),0)) + log(FB1(0,yl(xl),t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) -
		     log(FB2(0,0,t(s),0,tl(s6),b(c))) - log(FB2(0,yl(xl),t(s),0,tl(s6),0)) + log(FB2(0,yl(xl),t(s),0,tl(s6),b(c))))*pow(c,-1)*pow(mexp,-1)*pow(xl,-1)*XCfE0A0(s,0,s6);
}
double I4E0A0lc(double c, double xl, double s, double s6) {
	return 2*pow(c,-1)*pow(mexp,-1)*pow(xl,-1)*((mexp*log(xl)*(-log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB1(0,0,t(s),0,tl(s6),b(c))) + log(FB1(0,yl(xl),t(s),0,tl(s6),0)) - log(FB1(0,yl(xl),t(s),0,tl(s6),b(c))) -
	           log(FB2(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),b(c))) + log(FB2(0,yl(xl),t(s),0,tl(s6),0)) - log(FB2(0,yl(xl),t(s),0,tl(s6),b(c)))) +
	        8*log(c)*(log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) - log(FB1(0,yl(xl),t(s),0,tl(s6),0)) + log(FB1(0,yl(xl),t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) -
	           log(FB2(0,0,t(s),0,tl(s6),b(c))) - log(FB2(0,yl(xl),t(s),0,tl(s6),0)) + log(FB2(0,yl(xl),t(s),0,tl(s6),b(c)))) -
	        4*(pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(0,0,t(s),0,tl(s6),b(c))),2) - pow(log(FB1(0,yl(xl),t(s),0,tl(s6),0)),2) + pow(log(FB1(0,yl(xl),t(s),0,tl(s6),b(c))),2) +
	           pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(0,0,t(s),0,tl(s6),b(c))),2) - pow(log(FB2(0,yl(xl),t(s),0,tl(s6),0)),2) + pow(log(FB2(0,yl(xl),t(s),0,tl(s6),b(c))),2)))*XCfE0A0(s,0,s6) -
	     (log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) - log(FB2(0,0,t(s),0,tl(s6),b(c))))*(2*XCfE0A1(0,0,s,0,s6) + XCfE1A0(0,0,s,0,s6)) +
	     (log(FB1(0,yl(xl),t(s),0,tl(s6),0)) - log(FB1(0,yl(xl),t(s),0,tl(s6),b(c))) + log(FB2(0,yl(xl),t(s),0,tl(s6),0)) - log(FB2(0,yl(xl),t(s),0,tl(s6),b(c))))*(2*XCfE0A1(0,xl,s,0,s6) + XCfE1A0(0,xl,s,0,s6)));
}

double I4EM2A0ls(double xl, double s, double s5, double s6) {
	return 0; //leading order is xl-indep
}
double I4EM1A0ls(double xl, double s, double s5, double s6) {
	return (pow(mexp,-1)*pow(s5,-1)*pow(xl,-1)*(-2*XCfE0A1(0,0,s,0,s6) + 2*XCfE0A1(0,0,s,s5,s6) + 2*XCfE0A1(0,xl,s,0,s6) - 2*XCfE0A1(0,xl,s,s5,s6) + XCfE1A0(0,0,s,0,s6) - XCfE1A0(0,0,s,s5,s6) - XCfE1A0(0,xl,s,0,s6) +
		       XCfE1A0(0,xl,s,s5,s6)))/2.;
}
double I4E0A0ls(double xl, double s, double s5, double s6) {
	return (pow(mexp,-1)*pow(s5,-1)*pow(xl,-1)*(2*(mexp*log(xl)*(-log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB1(0,yl(xl),t(s),0,tl(s6),0)) - log(FB2(0,0,t(s),0,tl(s6),0)) + log(FB2(0,yl(xl),t(s),0,tl(s6),0))) -
	          2*(pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(0,yl(xl),t(s),0,tl(s6),0)),2) + pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(0,yl(xl),t(s),0,tl(s6),0)),2)))*XCfE0A0(s,0,s6) +
	       2*(mexp*log(xl)*(log(FB1(0,0,t(s),t5(s5),tl(s6),0)) - log(FB1(0,yl(xl),t(s),t5(s5),tl(s6),0)) + log(FB2(0,0,t(s),t5(s5),tl(s6),0)) - log(FB2(0,yl(xl),t(s),t5(s5),tl(s6),0))) +
	          2*(pow(log(FB1(0,0,t(s),t5(s5),tl(s6),0)),2) - pow(log(FB1(0,yl(xl),t(s),t5(s5),tl(s6),0)),2) + pow(log(FB2(0,0,t(s),t5(s5),tl(s6),0)),2) - pow(log(FB2(0,yl(xl),t(s),t5(s5),tl(s6),0)),2)))*
	        XCfE0A0(s,s5,s6) - 4*(log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),0)))*XCfE0A1(0,0,s,0,s6) + 4*(log(FB1(0,0,t(s),t5(s5),tl(s6),0)) + log(FB2(0,0,t(s),t5(s5),tl(s6),0)))*XCfE0A1(0,0,s,s5,s6) +
	       4*(log(FB1(0,yl(xl),t(s),0,tl(s6),0)) + log(FB2(0,yl(xl),t(s),0,tl(s6),0)))*XCfE0A1(0,xl,s,0,s6) -
	       4*(log(FB1(0,yl(xl),t(s),t5(s5),tl(s6),0)) + log(FB2(0,yl(xl),t(s),t5(s5),tl(s6),0)))*XCfE0A1(0,xl,s,s5,s6) +
	       mexp*log(xl)*(-XCfE1A0(0,0,s,0,s6) + XCfE1A0(0,0,s,s5,s6) + XCfE1A0(0,xl,s,0,s6) - XCfE1A0(0,xl,s,s5,s6)) +
	       2*log(s5)*(2*XCfE0A1(0,0,s,0,s6) - 2*XCfE0A1(0,0,s,s5,s6) - 2*XCfE0A1(0,xl,s,0,s6) + 2*XCfE0A1(0,xl,s,s5,s6) - XCfE1A0(0,0,s,0,s6) + XCfE1A0(0,0,s,s5,s6) + XCfE1A0(0,xl,s,0,s6) - XCfE1A0(0,xl,s,s5,s6)) -
	       2*XCfE1A1(0,0,s,0,s6) + 2*XCfE1A1(0,0,s,s5,s6) + 2*XCfE1A1(0,xl,s,0,s6) - 2*XCfE1A1(0,xl,s,s5,s6) + XCfE2A0(0,0,s,0,s6) - XCfE2A0(0,0,s,s5,s6) - XCfE2A0(0,xl,s,0,s6) + XCfE2A0(0,xl,s,s5,s6)))/2.;
}

double I4EM2A0cs(double c, double s, double s5, double s6) {
	return 0; //only alpha poles available, xk and xl are the only deltas
}
double I4EM1A0cs(double c, double s, double s5, double s6) {
	return 0; //only alpha poles available, xk and xl are the only deltas
}
double I4E0A0cs(double c, double s, double s5, double s6) {
	return 8*pow(c,-1)*pow(mexp,-2)*pow(s5,-1)*((2*log(c)*(-log(FB1(0,0,t(s),0,tl(s6),0)) + log(FB1(0,0,t(s),0,tl(s6),b(c))) - log(FB2(0,0,t(s),0,tl(s6),0)) + log(FB2(0,0,t(s),0,tl(s6),b(c)))) +
	        pow(log(FB1(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB1(0,0,t(s),0,tl(s6),b(c))),2) + pow(log(FB2(0,0,t(s),0,tl(s6),0)),2) - pow(log(FB2(0,0,t(s),0,tl(s6),b(c))),2))*XCfE0A0(s,0,s6) +
	     (2*log(c)*(log(FB1(0,0,t(s),t5(s5),tl(s6),0)) - log(FB1(0,0,t(s),t5(s5),tl(s6),b(c))) + log(FB2(0,0,t(s),t5(s5),tl(s6),0)) - log(FB2(0,0,t(s),t5(s5),tl(s6),b(c)))) -
	        pow(log(FB1(0,0,t(s),t5(s5),tl(s6),0)),2) + pow(log(FB1(0,0,t(s),t5(s5),tl(s6),b(c))),2) - pow(log(FB2(0,0,t(s),t5(s5),tl(s6),0)),2) + pow(log(FB2(0,0,t(s),t5(s5),tl(s6),b(c))),2))*XCfE0A0(s,s5,s6) +
	     (log(FB1(0,0,t(s),0,tl(s6),0)) - log(FB1(0,0,t(s),0,tl(s6),b(c))) + log(FB2(0,0,t(s),0,tl(s6),0)) - log(FB2(0,0,t(s),0,tl(s6),b(c))))*XCfE0A1(0,0,s,0,s6) +
	     (-log(FB1(0,0,t(s),t5(s5),tl(s6),0)) + log(FB1(0,0,t(s),t5(s5),tl(s6),b(c))) - log(FB2(0,0,t(s),t5(s5),tl(s6),0)) + log(FB2(0,0,t(s),t5(s5),tl(s6),b(c))))*XCfE0A1(0,0,s,s5,s6));
}

